//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module simply pipelines the stream and only allows the next packet to flow thru the pipeline
//	when it has sufficient full packet credits to allow only full packets to pass thru the pipeline.
//
//	EOP_CREDIT_WIDTH should be set such that EOP_CREDIT_LIMIT = 2 ** ( EOP_CREDIT_WIDTH - 1 ) will
//	provide enough counter width to increment thru the highest anticipated credit count.
//	

module full_packet_wait_gate_64b
#(
	parameter EOP_CREDIT_WIDTH = 6
)
(
	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// Input Interface - Avalon ST
	input  wire	[ 63 : 0 ]	asi_snk_data,
	output wire				asi_snk_ready,
	input  wire				asi_snk_valid,
	input  wire				asi_snk_startofpacket,
	input  wire				asi_snk_endofpacket,

	// Output Interface - Avalon ST
	output wire	[ 63 : 0 ]	aso_src_data,
	input  wire				aso_src_ready,
	output wire				aso_src_valid,
	output wire				aso_src_startofpacket,
	output wire				aso_src_endofpacket,
	
	// Credit Increment Report Interface - Avalon ST
	input  wire				asi_eopcreditincr_data
);

reg  [ ( EOP_CREDIT_WIDTH - 1 ) : 0 ] eop_credit;
wire incr_credit;
wire decr_credit;

assign incr_credit = asi_eopcreditincr_data;
assign decr_credit = asi_snk_ready & asi_snk_valid & asi_snk_endofpacket;

always @ ( posedge clk or posedge reset ) begin

	if( reset ) begin
	
		eop_credit <= 0;
		
	end else begin
		
		eop_credit <= eop_credit + incr_credit - decr_credit;
	end
end

wire in_ready;
wire in_valid;

assign asi_snk_ready =	( asi_snk_startofpacket ) ?
						( eop_credit > 0 ) ? ( in_ready ) : ( 1'b0 ) :
						( in_ready );

assign in_valid =	( asi_snk_startofpacket ) ?
					( eop_credit > 0 ) ? ( asi_snk_valid ) : ( 1'b0 ) :
					( asi_snk_valid );

//
// readdata output Pipeline
//
full_packet_wait_gate_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 66 ) ) outpipe (
    .clk            ( clk ),
    .reset_n        ( ~reset ),
    .in_ready       ( in_ready ),
    .in_valid       ( in_valid ), 
    .in_payload     ( { asi_snk_data, asi_snk_startofpacket, asi_snk_endofpacket } ),
    .out_ready      ( aso_src_ready ), 
    .out_valid      ( aso_src_valid ), 
    .out_payload    ( { aso_src_data, aso_src_startofpacket, aso_src_endofpacket } )
);

endmodule

//  --------------------------------------------------------------------------------
// | single buffered pipeline stage
//  --------------------------------------------------------------------------------
module full_packet_wait_gate_64b_1stage_pipeline  
#( parameter PAYLOAD_WIDTH = 8 )
 ( input                          clk,
   input                          reset_n, 
   output reg                     in_ready,
   input                          in_valid,   
   input      [PAYLOAD_WIDTH-1:0] in_payload,
   input                          out_ready,   
   output reg                     out_valid,
   output reg [PAYLOAD_WIDTH-1:0] out_payload      
 );
      
   always @* begin
     in_ready <= out_ready || ~out_valid;
   end
   
   always @ (negedge reset_n, posedge clk) begin
      if (!reset_n) begin
         out_valid <= 0;
         out_payload <= 0;
      end else begin
         if (in_valid) begin
           out_valid <= 1;
         end else if (out_ready) begin
           out_valid <= 0;
         end
         
         if(in_valid && in_ready) begin
            out_payload <= in_payload;
         end
      end
   end

endmodule
