//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module provides a simple interface to the MSI functionality of the PCIe HIP core.
//
//  There is a dummy master interface on this peripheral that facilited the master association required by the
//  IRQ receiver interface.  Just connect this master to any slave like the CSR slave in the SOPC system.
//	
//	Register maps of Avalon Slave CSR
//
//	'h00: [ 31 : 0 ] { 7'h00, mask_capable, addr_64_capable, mult_mess_enable, mult_mess_capable, msi_enable, 13'h00, tc_reg }
//			RW : [  2 : 0 ]	tc_reg				// traffic class for MSI packet
//			RO : msi_enable						// MSI enable status bit from HIP core
//			RO : [ 2 : 0 ]	mult_mess_capable	// Multiple Message Capable status from HIP core
//			RO : [ 2 : 0 ]	mult_mess_enable	// Multiple Message Allocated status from HIP core
//			RO : addr_64_capable				// 64-bit address capable status from HIP core
//			RO : mask_capable					// MSI Maskable status from HIP core
//	'h01: RW : [ 31 : 0 ] irq_enable_reg		// each bit in this register enables the corresponding IRQ input
//	'h02: RO : [ 31 : 0 ] sync_inr_irqin_irq	// this is the live state of each IRQ input
//	'h03: RO : [ 31 : 0 ] msi_sent_reg			// this register indicates that an MSI message has been sent for the corresponding IRQ
//												// these bits are set when the MSI is sent and they are cleared when the IRQ is deactivated at its source
//
module msi_interface
(
	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// Avalon Master Interface - Avalon MM - this is a dummy master to facilitate the IRQ receiver
	output wire	[ 31 : 0 ]	avm_msi_master_address,
	output wire	[  3 : 0 ]	avm_msi_master_byteenable,
	output wire				avm_msi_master_read,
	output wire				avm_msi_master_write,
	input  wire				avm_msi_master_waitrequest,
	output wire	[ 31 : 0 ]	avm_msi_master_writedata,
	input  wire	[ 31 : 0 ]	avm_msi_master_readdata,
	
	// Avalon Slave Interface - Avalon MM
	input  wire	[  1 : 0 ]	avs_msi_slave_address,
	input  wire	[  3 : 0 ]	avs_msi_slave_byteenable,
	input  wire				avs_msi_slave_read,
	input  wire				avs_msi_slave_write,
	output wire				avs_msi_slave_waitrequest,
	input  wire	[ 31 : 0 ]	avs_msi_slave_writedata,
	output wire	[ 31 : 0 ]	avs_msi_slave_readdata,
	
	// Avalon IRQ Receiver Interface - Avalon MM
	input  wire	[ 31 : 0 ]	inr_irqin_irq,

	// PCIe MSICSR interface - Avalon ST
	input  wire [ 15 : 0 ]	asi_msicsr_data,
	
	// MSI Interface
	output wire 			coe_msi_req,
	input  wire 			coe_msi_ack,
	output wire [  2 : 0 ]	coe_msi_tc,
	output wire [  4 : 0 ]	coe_msi_num
);

wire 			msi_enable;
wire [ 2 : 0 ]	mult_mess_capable;
wire [ 2 : 0 ]	mult_mess_enable;
wire			addr_64_capable;
wire			mask_capable;
wire [ 6 : 0 ]	reserved;

wire [ 31 : 0 ]	allocated_msi_mask;
wire [ 31 : 0 ]	outstanding_irqs;
wire [ 31 : 0 ]	next_msi_to_send;
reg  [ 31 : 0 ]	next_msi_to_send_reg;
reg 			msi_req;

reg  [  2 : 0 ]	tc_reg;
reg  [ 31 : 0 ]	irq_enable_reg;
reg  [ 31 : 0 ]	msi_sent_reg;

wire [ 31 : 0 ]	sync_inr_irqin_irq;

altera_std_synchronizer_bundle
#
(
    .width	( 32 ),
    .depth	( 3 )
)
altera_std_synchronizer_bundle_inst
(
	.clk		( clk ),
	.reset_n	( ~reset ),
	.din		( inr_irqin_irq ),
	.dout		( sync_inr_irqin_irq )
);

assign coe_msi_req = msi_req;

assign coe_msi_tc = tc_reg;

assign { reserved, mask_capable, addr_64_capable, mult_mess_enable, mult_mess_capable, msi_enable } = asi_msicsr_data;

assign allocated_msi_mask	=	( mult_mess_enable == 3'h0 ) ? ( 32'h00000001 ) :
								( mult_mess_enable == 3'h1 ) ? ( 32'h00000003 ) :
								( mult_mess_enable == 3'h2 ) ? ( 32'h0000000F ) :
								( mult_mess_enable == 3'h3 ) ? ( 32'h000000FF ) :
								( mult_mess_enable == 3'h4 ) ? ( 32'h0000FFFF ) :
								( mult_mess_enable == 3'h5 ) ? ( 32'hFFFFFFFF ) :
															   ( 32'h00000000 );

assign avs_msi_slave_waitrequest	=	0;

assign avs_msi_slave_readdata		=	( avs_msi_slave_address == 0 ) ? ( { 7'h00, mask_capable, addr_64_capable, mult_mess_enable, mult_mess_capable, msi_enable, 13'h00, tc_reg } ) :
										( avs_msi_slave_address == 1 ) ? ( irq_enable_reg ) :
										( avs_msi_slave_address == 2 ) ? ( sync_inr_irqin_irq ) :
																		 ( msi_sent_reg );

assign outstanding_irqs = ( allocated_msi_mask & irq_enable_reg & sync_inr_irqin_irq & ~msi_sent_reg );

assign next_msi_to_send = 	( msi_enable ) ? 
								( outstanding_irqs[0] ) ? ( 32'h00000001 ) :
								( outstanding_irqs[1] ) ? ( 32'h00000002 ) :
								( outstanding_irqs[2] ) ? ( 32'h00000004 ) :
								( outstanding_irqs[3] ) ? ( 32'h00000008 ) :
								( outstanding_irqs[4] ) ? ( 32'h00000010 ) :
								( outstanding_irqs[5] ) ? ( 32'h00000020 ) :
								( outstanding_irqs[6] ) ? ( 32'h00000040 ) :
								( outstanding_irqs[7] ) ? ( 32'h00000080 ) :
								( outstanding_irqs[8] ) ? ( 32'h00000100 ) :
								( outstanding_irqs[9] ) ? ( 32'h00000200 ) :
								( outstanding_irqs[10] ) ? ( 32'h00000400 ) :
								( outstanding_irqs[11] ) ? ( 32'h00000800 ) :
								( outstanding_irqs[12] ) ? ( 32'h00001000 ) :
								( outstanding_irqs[13] ) ? ( 32'h00002000 ) :
								( outstanding_irqs[14] ) ? ( 32'h00004000 ) :
								( outstanding_irqs[15] ) ? ( 32'h00008000 ) :
								( outstanding_irqs[16] ) ? ( 32'h00010000 ) :
								( outstanding_irqs[17] ) ? ( 32'h00020000 ) :
								( outstanding_irqs[18] ) ? ( 32'h00040000 ) :
								( outstanding_irqs[19] ) ? ( 32'h00080000 ) :
								( outstanding_irqs[20] ) ? ( 32'h00100000 ) :
								( outstanding_irqs[21] ) ? ( 32'h00200000 ) :
								( outstanding_irqs[22] ) ? ( 32'h00400000 ) :
								( outstanding_irqs[23] ) ? ( 32'h00800000 ) :
								( outstanding_irqs[24] ) ? ( 32'h01000000 ) :
								( outstanding_irqs[25] ) ? ( 32'h02000000 ) :
								( outstanding_irqs[26] ) ? ( 32'h04000000 ) :
								( outstanding_irqs[27] ) ? ( 32'h08000000 ) :
								( outstanding_irqs[28] ) ? ( 32'h10000000 ) :
								( outstanding_irqs[29] ) ? ( 32'h20000000 ) :
								( outstanding_irqs[30] ) ? ( 32'h40000000 ) :
								( outstanding_irqs[31] ) ? ( 32'h80000000 ) :
														   ( 32'h00000000 )
								: ( 32'h00000000 );

assign coe_msi_num =	( next_msi_to_send_reg[0] ) ?  ( 5'h00 ) :
						( next_msi_to_send_reg[1] ) ?  ( 5'h01 ) :
						( next_msi_to_send_reg[2] ) ?  ( 5'h02 ) :
						( next_msi_to_send_reg[3] ) ?  ( 5'h03 ) :
						( next_msi_to_send_reg[4] ) ?  ( 5'h04 ) :
						( next_msi_to_send_reg[5] ) ?  ( 5'h05 ) :
						( next_msi_to_send_reg[6] ) ?  ( 5'h06 ) :
						( next_msi_to_send_reg[7] ) ?  ( 5'h07 ) :
						( next_msi_to_send_reg[8] ) ?  ( 5'h08 ) :
						( next_msi_to_send_reg[9] ) ?  ( 5'h09 ) :
						( next_msi_to_send_reg[10] ) ? ( 5'h0A ) :
						( next_msi_to_send_reg[11] ) ? ( 5'h0B ) :
						( next_msi_to_send_reg[12] ) ? ( 5'h0C ) :
						( next_msi_to_send_reg[13] ) ? ( 5'h0D ) :
						( next_msi_to_send_reg[14] ) ? ( 5'h0E ) :
						( next_msi_to_send_reg[15] ) ? ( 5'h0F ) :
						( next_msi_to_send_reg[16] ) ? ( 5'h10 ) :
						( next_msi_to_send_reg[17] ) ? ( 5'h11 ) :
						( next_msi_to_send_reg[18] ) ? ( 5'h12 ) :
						( next_msi_to_send_reg[19] ) ? ( 5'h13 ) :
						( next_msi_to_send_reg[20] ) ? ( 5'h14 ) :
						( next_msi_to_send_reg[21] ) ? ( 5'h15 ) :
						( next_msi_to_send_reg[22] ) ? ( 5'h16 ) :
						( next_msi_to_send_reg[23] ) ? ( 5'h17 ) :
						( next_msi_to_send_reg[24] ) ? ( 5'h18 ) :
						( next_msi_to_send_reg[25] ) ? ( 5'h19 ) :
						( next_msi_to_send_reg[26] ) ? ( 5'h1A ) :
						( next_msi_to_send_reg[27] ) ? ( 5'h1B ) :
						( next_msi_to_send_reg[28] ) ? ( 5'h1C ) :
						( next_msi_to_send_reg[29] ) ? ( 5'h1D ) :
						( next_msi_to_send_reg[30] ) ? ( 5'h1E ) :
						( next_msi_to_send_reg[31] ) ? ( 5'h1F ) :
												       ( 5'h00 );

//
// slave write state machine
//
always @ ( posedge clk or posedge reset ) begin

	if( reset ) begin

		tc_reg         <= 0;
		irq_enable_reg <= 0;
		
	end else begin

		if( avs_msi_slave_write ) begin
			
			case( avs_msi_slave_address )
			2'h0: begin
				if( avs_msi_slave_byteenable[0] ) begin
					tc_reg <= avs_msi_slave_writedata[ 2 : 0 ];
				end
			end
			2'h1: begin
				if( avs_msi_slave_byteenable[0] ) begin
					irq_enable_reg[  7 :  0 ] <= avs_msi_slave_writedata[  7 :  0 ];
				end
				if( avs_msi_slave_byteenable[1] ) begin
					irq_enable_reg[ 15 :  8 ] <= avs_msi_slave_writedata[ 15 :  8 ];
				end
				if( avs_msi_slave_byteenable[2] ) begin
					irq_enable_reg[ 23 : 16 ] <= avs_msi_slave_writedata[ 23 : 16 ];
				end
				if( avs_msi_slave_byteenable[3] ) begin
					irq_enable_reg[ 31 : 24 ] <= avs_msi_slave_writedata[ 31 : 24 ];
				end
			end
			default: begin
			end
			endcase
			
		end
	end
end

//
// msi request state machine
//
always @ ( posedge clk or posedge reset ) begin

	if( reset ) begin

		next_msi_to_send_reg    <= 0;
		msi_req					<= 0;
		msi_sent_reg			<= 0;
		
	end else begin

		if( ~msi_req ) begin

			if( |next_msi_to_send_reg ) begin

				msi_req					<= 1;
				
			end else begin

				next_msi_to_send_reg	<= next_msi_to_send;
			end
		end else begin
			if( coe_msi_ack ) begin

				next_msi_to_send_reg    <= 0;
				msi_req					<= 0;
			end
		end

		if( msi_req & coe_msi_ack ) begin
			case( next_msi_to_send_reg )
			32'h00000001: msi_sent_reg[0]  <= sync_inr_irqin_irq[0];
			32'h00000002: msi_sent_reg[1]  <= sync_inr_irqin_irq[1];
			32'h00000004: msi_sent_reg[2]  <= sync_inr_irqin_irq[2];
			32'h00000008: msi_sent_reg[3]  <= sync_inr_irqin_irq[3];
			32'h00000010: msi_sent_reg[4]  <= sync_inr_irqin_irq[4];
			32'h00000020: msi_sent_reg[5]  <= sync_inr_irqin_irq[5];
			32'h00000040: msi_sent_reg[6]  <= sync_inr_irqin_irq[6];
			32'h00000080: msi_sent_reg[7]  <= sync_inr_irqin_irq[7];
			32'h00000100: msi_sent_reg[8]  <= sync_inr_irqin_irq[8];
			32'h00000200: msi_sent_reg[9]  <= sync_inr_irqin_irq[9];
			32'h00000400: msi_sent_reg[10] <= sync_inr_irqin_irq[10];
			32'h00000800: msi_sent_reg[11] <= sync_inr_irqin_irq[11];
			32'h00001000: msi_sent_reg[12] <= sync_inr_irqin_irq[12];
			32'h00002000: msi_sent_reg[13] <= sync_inr_irqin_irq[13];
			32'h00004000: msi_sent_reg[14] <= sync_inr_irqin_irq[14];
			32'h00008000: msi_sent_reg[15] <= sync_inr_irqin_irq[15];
			32'h00010000: msi_sent_reg[16] <= sync_inr_irqin_irq[16];
			32'h00020000: msi_sent_reg[17] <= sync_inr_irqin_irq[17];
			32'h00040000: msi_sent_reg[18] <= sync_inr_irqin_irq[18];
			32'h00080000: msi_sent_reg[19] <= sync_inr_irqin_irq[19];
			32'h00100000: msi_sent_reg[20] <= sync_inr_irqin_irq[20];
			32'h00200000: msi_sent_reg[21] <= sync_inr_irqin_irq[21];
			32'h00400000: msi_sent_reg[22] <= sync_inr_irqin_irq[22];
			32'h00800000: msi_sent_reg[23] <= sync_inr_irqin_irq[23];
			32'h01000000: msi_sent_reg[24] <= sync_inr_irqin_irq[24];
			32'h02000000: msi_sent_reg[25] <= sync_inr_irqin_irq[25];
			32'h04000000: msi_sent_reg[26] <= sync_inr_irqin_irq[26];
			32'h08000000: msi_sent_reg[27] <= sync_inr_irqin_irq[27];
			32'h10000000: msi_sent_reg[28] <= sync_inr_irqin_irq[28];
			32'h20000000: msi_sent_reg[29] <= sync_inr_irqin_irq[29];
			32'h40000000: msi_sent_reg[30] <= sync_inr_irqin_irq[30];
			32'h80000000: msi_sent_reg[31] <= sync_inr_irqin_irq[31];
			default: begin
			end
			endcase
		end else begin
			if( msi_sent_reg[0] )  msi_sent_reg[0]  <= sync_inr_irqin_irq[0];
			if( msi_sent_reg[1] )  msi_sent_reg[1]  <= sync_inr_irqin_irq[1];
			if( msi_sent_reg[2] )  msi_sent_reg[2]  <= sync_inr_irqin_irq[2];
			if( msi_sent_reg[3] )  msi_sent_reg[3]  <= sync_inr_irqin_irq[3];
			if( msi_sent_reg[4] )  msi_sent_reg[4]  <= sync_inr_irqin_irq[4];
			if( msi_sent_reg[5] )  msi_sent_reg[5]  <= sync_inr_irqin_irq[5];
			if( msi_sent_reg[6] )  msi_sent_reg[6]  <= sync_inr_irqin_irq[6];
			if( msi_sent_reg[7] )  msi_sent_reg[7]  <= sync_inr_irqin_irq[7];
			if( msi_sent_reg[8] )  msi_sent_reg[8]  <= sync_inr_irqin_irq[8];
			if( msi_sent_reg[9] )  msi_sent_reg[9]  <= sync_inr_irqin_irq[9];
			if( msi_sent_reg[10] ) msi_sent_reg[10] <= sync_inr_irqin_irq[10];
			if( msi_sent_reg[11] ) msi_sent_reg[11] <= sync_inr_irqin_irq[11];
			if( msi_sent_reg[12] ) msi_sent_reg[12] <= sync_inr_irqin_irq[12];
			if( msi_sent_reg[13] ) msi_sent_reg[13] <= sync_inr_irqin_irq[13];
			if( msi_sent_reg[14] ) msi_sent_reg[14] <= sync_inr_irqin_irq[14];
			if( msi_sent_reg[15] ) msi_sent_reg[15] <= sync_inr_irqin_irq[15];
			if( msi_sent_reg[16] ) msi_sent_reg[16] <= sync_inr_irqin_irq[16];
			if( msi_sent_reg[17] ) msi_sent_reg[17] <= sync_inr_irqin_irq[17];
			if( msi_sent_reg[18] ) msi_sent_reg[18] <= sync_inr_irqin_irq[18];
			if( msi_sent_reg[19] ) msi_sent_reg[19] <= sync_inr_irqin_irq[19];
			if( msi_sent_reg[20] ) msi_sent_reg[20] <= sync_inr_irqin_irq[20];
			if( msi_sent_reg[21] ) msi_sent_reg[21] <= sync_inr_irqin_irq[21];
			if( msi_sent_reg[22] ) msi_sent_reg[22] <= sync_inr_irqin_irq[22];
			if( msi_sent_reg[23] ) msi_sent_reg[23] <= sync_inr_irqin_irq[23];
			if( msi_sent_reg[24] ) msi_sent_reg[24] <= sync_inr_irqin_irq[24];
			if( msi_sent_reg[25] ) msi_sent_reg[25] <= sync_inr_irqin_irq[25];
			if( msi_sent_reg[26] ) msi_sent_reg[26] <= sync_inr_irqin_irq[26];
			if( msi_sent_reg[27] ) msi_sent_reg[27] <= sync_inr_irqin_irq[27];
			if( msi_sent_reg[28] ) msi_sent_reg[28] <= sync_inr_irqin_irq[28];
			if( msi_sent_reg[29] ) msi_sent_reg[29] <= sync_inr_irqin_irq[29];
			if( msi_sent_reg[30] ) msi_sent_reg[30] <= sync_inr_irqin_irq[30];
			if( msi_sent_reg[31] ) msi_sent_reg[31] <= sync_inr_irqin_irq[31];
		end
	end
end

//
// Terminate the dummy master outputs
//
assign avm_msi_master_address		= 0;
assign avm_msi_master_byteenable	= 0;
assign avm_msi_master_read			= 0;
assign avm_msi_master_write			= 0;
assign avm_msi_master_writedata		= 0;

endmodule
