//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module receives 3DW(32-bit) and 4DW(64-bit) request TLPs and converts them into Avalon MM
//	Master commands as well as read completion information streams.  This module manages a read data
//	credit scheme such that it does not oversubscribe the external read data FIFO which is described by
//	the READ_CREDIT_WIDTH parameter.  You should set READ_CREDIT_WIDTH such that the READ_CREDIT_LIMIT
//	macro is calculated properly by this formula: READ_CREDIT_LIMIT = 2 ** ( READ_CREDIT_WIDTH - 1 ).
//	READ_CREDIT_LIMIT should not exceed the total number of Avalon ST data words that the external read
//	data FIFO can store.
//
//	The AVALON_MASTER_ADDRESS_WIDTH parameter should be set to whatever width you require on the Avalon
//	MM master that drives the ultimate transactions into the system.
//	

module pcie_64bar_tlp2master_64b_piped
#(
	parameter AVALON_MASTER_ADDRESS_WIDTH = 16,
	parameter COMPLETION_INFO_WIDTH = 64,
	parameter READ_CREDIT_WIDTH = 10
)
(

	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// TLP Input Interface - PCIe ST
	input  wire	[ 63 : 0 ]	asi_tlpsnk_data,
	output wire				asi_tlpsnk_ready,
	input  wire				asi_tlpsnk_valid,
	input  wire				asi_tlpsnk_startofpacket,
	input  wire				asi_tlpsnk_endofpacket,
	
	// Avalon Master Command Output Interface - Avalon ST
	output wire	[ ( AVALON_MASTER_ADDRESS_WIDTH + 83 - 1 ) : 0 ]	aso_cmdsrc_data,
	input  wire				aso_cmdsrc_ready,
	output wire				aso_cmdsrc_valid,

	// Read Completion Info Output Interface - Avalon ST
	output wire	[ ( COMPLETION_INFO_WIDTH - 1 ) : 0 ]	aso_cplinfosrc_data,
	input  wire				aso_cplinfosrc_ready,
	output wire				aso_cplinfosrc_valid,

	// Read Credit Decrement Input Interface - Avalon ST
	input  wire 			asi_readcreditdec_data
	
);

localparam READ_CREDIT_LIMIT = 2 ** ( READ_CREDIT_WIDTH - 1 );

//
// Things we need to keep out of the TLP header
//
reg  [  1 : 0 ] request_fmt;
reg  [  2 : 0 ] request_tc;
reg  [  1 : 0 ] request_attr;
reg  [  9 : 0 ] request_length;
reg  [ 15 : 0 ] request_reqid;
reg  [  7 : 0 ] request_tag;
reg  [  3 : 0 ] request_lbe;
reg  [  3 : 0 ] request_fbe;

//
// General packet information
//
reg				sop_plus_one;
reg				sop_plus_two;

reg				request_format_error;

//
// Things we need to compute
//
wire [ ( AVALON_MASTER_ADDRESS_WIDTH - 1 ) : 0 ] master_address_field;
reg  [ ( AVALON_MASTER_ADDRESS_WIDTH - 1 ) : 0 ] reg_master_address;
reg  [ 10 : 0 ]	regp_request_length_plus_1;
reg  [ 10 : 0 ]	regp_request_length_plus_2;
wire [  9 : 0 ]	next_master_burstcount;
reg  [  9 : 0 ]	reg_master_burstcount;
reg  [ ( READ_CREDIT_WIDTH - 1 ) : 0] read_credits;
//wire			read_credit_available;
wire [  9 : 0 ]	credits_to_add;
wire [  9 : 0 ]	credits_to_sub;
reg  [  9 : 0 ]	regp_credits_to_add;
reg  [  9 : 0 ]	regp_credits_to_sub;

//
// Master Command Vector Elements
//
wire [ ( AVALON_MASTER_ADDRESS_WIDTH - 1 ) : 0 ] avm_address;
wire [  7 : 0 ]	avm_byteenable;
wire [  7 : 0 ]	unaligned_write_3dw_be;
wire [  7 : 0 ]	all_other_byteenable;
wire 			avm_write;
wire [ 63 : 0 ]	avm_writedata;
wire [  9 : 0 ] avm_burstcount;
wire [ ( AVALON_MASTER_ADDRESS_WIDTH + 83 - 1 ) : 0 ] avm_command;

wire			command_outpipe_ready;
wire			command_outpipe_valid;

wire			cplinfo_outpipe_ready;
wire			cplinfo_outpipe_valid;

wire 			read_completions_ready;

//
// Completion Info Elements
//
wire [ ( COMPLETION_INFO_WIDTH - 1 ) : 0 ]	completion_info;
assign completion_info =	{
								request_attr,
								request_tc,
								request_tag,
								request_reqid,
								request_lbe,
								request_fbe,
								request_length,
								master_address_field[  6 : 0 ],
								next_master_burstcount
							};

//
// TLP header word 0
//
wire [  0 : 0 ] tlp_res3_field;
wire [  1 : 0 ] tlp_fmt_field;
wire [  4 : 0 ] tlp_type_field;
wire [  0 : 0 ] tlp_res2_field;
wire [  2 : 0 ] tlp_tc_field;
wire [  3 : 0 ] tlp_res1_field;
wire [  0 : 0 ] tlp_td_field;
wire [  0 : 0 ] tlp_ep_field;
wire [  1 : 0 ] tlp_attr_field;
wire [  1 : 0 ] tlp_res0_field;
wire [  9 : 0 ] tlp_length_field;

//
// TLP header word 1
//
wire [ 15 : 0 ] tlp_reqid_field;
wire [  7 : 0 ] tlp_tag_field;
wire [  3 : 0 ] tlp_lbe_field;
wire [  3 : 0 ] tlp_fbe_field;

//
// TLP header word 2 and 3
//
wire [ 61 : 0 ] tlp_address_field;
wire [  1 : 0 ] tlp_res4_field;

assign {
			tlp_reqid_field,
			tlp_tag_field,
			tlp_lbe_field,
			tlp_fbe_field,
			tlp_res3_field,
			tlp_fmt_field,
			tlp_type_field,
			tlp_res2_field,
			tlp_tc_field,
			tlp_res1_field,
			tlp_td_field,
			tlp_ep_field,
			tlp_attr_field,
			tlp_res0_field,
			tlp_length_field
							} = asi_tlpsnk_data;

assign {
			tlp_address_field,
			tlp_res4_field
							} = ( request_fmt[0] == 1 ) ? ( { asi_tlpsnk_data[ 31 : 0 ], asi_tlpsnk_data[ 63 : 32 ] } ) : ( { 32'h0, asi_tlpsnk_data[ 31 : 0 ] } );

assign next_master_burstcount =	( request_length == 10'd1 ) ?
									( 10'd1 ) :
									( tlp_address_field[0] ) ? 
										( regp_request_length_plus_2[ 10 : 1 ] ) : 
										( regp_request_length_plus_1[ 10 : 1 ] ); // tlp_address_field[0] = a[2]

assign master_address_field = { tlp_address_field[ 61 : 0 ], 2'b00 };

assign avm_address = ( sop_plus_one ) ? ( master_address_field ) : ( reg_master_address );

assign avm_byteenable =	( ( request_fmt[1] == 1 ) && ( request_fmt[0] == 0 ) && ( master_address_field[2] == 1 ) && ( sop_plus_one ) ||
						  ( request_fmt[1] == 1 ) && ( request_fmt[0] == 0 ) && ( reg_master_address[2] == 1 ) && ( !sop_plus_one ) ) ?
								( unaligned_write_3dw_be ) :
								( all_other_byteenable );
									
assign unaligned_write_3dw_be =	( sop_plus_one ) ?
									( { request_fbe, 4'h0 } ) :
									( asi_tlpsnk_endofpacket == 0 ) ?
										( 8'hFF ) :
										( request_length[0] == 0 ) ? 
												( { 4'h0, request_lbe } ) : 
												( { request_lbe, 4'hF } ); 

assign all_other_byteenable =	( request_fmt[1] == 0 ) ?
									( request_length == 10'd1 ) ?
										( master_address_field[2] == 0 ) ?
											( { 4'h0, request_fbe } ) :
											( { request_fbe, 4'h0 } ) :
										( 8'hFF ) :
									( request_length == 10'd1 ) ?
										( reg_master_address[2] == 0 ) ? 
											( { 4'h0, request_fbe } ) : 
											( { request_fbe, 4'h0 } ) :
										( ( request_length == 10'd2 ) && ( reg_master_address[2] == 0 ) && ( sop_plus_two == 1 ) ) ?
											( { request_lbe, request_fbe } ) :
											( sop_plus_two == 1 ) ?
												( reg_master_address[2] == 0  ) ? 
													( { 4'hF, request_fbe } ) : 
													( { request_fbe, 4'h0 } ) :
												( asi_tlpsnk_endofpacket == 0 ) ?
													( 8'hFF ) :
													( request_length[0] == 0 ) ? 
														( reg_master_address[2] == 0 ) ? 
															( { request_lbe, 4'hF } ) : 
															( { 4'h0, request_lbe } ) : 
														( reg_master_address[2] == 0 ) ? 
															( { 4'h0, request_lbe } ) : 
															( { request_lbe, 4'hF } ); 

assign avm_burstcount = ( sop_plus_one ) ? ( next_master_burstcount ) : ( reg_master_burstcount );

assign avm_write = request_fmt[1];

assign avm_writedata = asi_tlpsnk_data;

assign avm_command = { avm_address, avm_byteenable, avm_burstcount, avm_write, avm_writedata };

//
// Avalon ST handshaking
//
assign asi_tlpsnk_ready = 	( asi_tlpsnk_startofpacket ) | 
							( sop_plus_one & command_outpipe_ready & read_completions_ready ) |
							( ~asi_tlpsnk_startofpacket & ~sop_plus_one & command_outpipe_ready );

assign command_outpipe_valid =	( sop_plus_one & ~request_fmt[1] & asi_tlpsnk_valid & asi_tlpsnk_ready ) |
								( sop_plus_one & ~request_fmt[0] & master_address_field[2] & asi_tlpsnk_valid & asi_tlpsnk_ready ) |
								( ~asi_tlpsnk_startofpacket & ~sop_plus_one & asi_tlpsnk_valid & asi_tlpsnk_ready );

assign read_completions_ready =	( sop_plus_one ) ?
									( request_fmt[1] == 1'b1 ) ?
										( 1'b1 ) :
										( cplinfo_outpipe_ready == 1'b0 ) ? 
											( 1'b0 ) :
											( ( ( regp_request_length_plus_2[ 10 : 1 ] ) + read_credits ) > READ_CREDIT_LIMIT ) ? 
												( 1'b0 ) : 
												( 1'b1 ) :
									( 1'b0 );

assign cplinfo_outpipe_valid =	( sop_plus_one & ~request_fmt[1] & asi_tlpsnk_valid & asi_tlpsnk_ready );

//
// TLP header capture
//
always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin

		request_fmt				<= 0;
		request_tc				<= 0;
		request_attr			<= 0;
		request_length			<= 0;
		regp_request_length_plus_1 <= 0;
		regp_request_length_plus_2 <= 0;
		request_reqid			<= 0;
		request_tag				<= 0;
		request_lbe				<= 0;
		request_fbe				<= 0;

		sop_plus_one			<= 0;
		sop_plus_two			<= 0;
		reg_master_address		<= 0;
		reg_master_burstcount	<= 0;

	end else begin

		if( asi_tlpsnk_ready & asi_tlpsnk_valid ) begin

			if( asi_tlpsnk_startofpacket ) begin
			
				request_fmt		<= tlp_fmt_field;
				request_tc		<= tlp_tc_field;
				request_attr	<= tlp_attr_field;
				request_length	<= tlp_length_field;
				regp_request_length_plus_1 <= ( tlp_length_field == 0 ) ? ( 11'h401 ) : ( tlp_length_field + 1 );
				regp_request_length_plus_2 <= ( tlp_length_field == 0 ) ? ( 11'h402 ) : ( tlp_length_field + 2 );
				request_reqid	<= tlp_reqid_field;
				request_tag		<= tlp_tag_field;
				request_lbe		<= tlp_lbe_field;
				request_fbe		<= tlp_fbe_field;
			end
			
			sop_plus_one	<= asi_tlpsnk_startofpacket;
			sop_plus_two	<= sop_plus_one;
			
			if( sop_plus_one ) begin
			
				reg_master_address		<= master_address_field;
				reg_master_burstcount	<= next_master_burstcount;
			end
		end	
	end
end

//
// Read Credit Managment
//
assign credits_to_add = ( command_outpipe_valid & command_outpipe_ready & ~avm_write ) ? ( avm_burstcount ) : ( 10'd0 );

assign credits_to_sub = ( asi_readcreditdec_data ) ? ( 10'd1 ) : ( 10'd0 );

always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin

		read_credits <= 0;
		regp_credits_to_add <= 0;
		regp_credits_to_sub <= 0;
		
	end else begin

		read_credits <= read_credits + regp_credits_to_add - regp_credits_to_sub;
		regp_credits_to_add <= credits_to_add;
		regp_credits_to_sub <= credits_to_sub;
		
	end
end

//
// command output Pipeline
//
pcie_64bar_tlp2master_64b_piped_1stage_pipeline #( .PAYLOAD_WIDTH( AVALON_MASTER_ADDRESS_WIDTH + 83 ) ) command_outpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (command_outpipe_ready),
    .in_valid       (command_outpipe_valid), 
    .in_payload     (avm_command),
    .out_ready      (aso_cmdsrc_ready), 
    .out_valid      (aso_cmdsrc_valid), 
    .out_payload    (aso_cmdsrc_data)
);

//
// completion info output Pipeline
//
pcie_64bar_tlp2master_64b_piped_1stage_pipeline #( .PAYLOAD_WIDTH( COMPLETION_INFO_WIDTH ) ) completion_info_outpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (cplinfo_outpipe_ready),
    .in_valid       (cplinfo_outpipe_valid), 
    .in_payload     (completion_info),
    .out_ready      (aso_cplinfosrc_ready), 
    .out_valid      (aso_cplinfosrc_valid), 
    .out_payload    (aso_cplinfosrc_data)
);

endmodule

//  --------------------------------------------------------------------------------
// | single buffered pipeline stage
//  --------------------------------------------------------------------------------
module pcie_64bar_tlp2master_64b_piped_1stage_pipeline  
#( parameter PAYLOAD_WIDTH = 8 )
 ( input                          clk,
   input                          reset_n, 
   output reg                     in_ready,
   input                          in_valid,   
   input      [PAYLOAD_WIDTH-1:0] in_payload,
   input                          out_ready,   
   output reg                     out_valid,
   output reg [PAYLOAD_WIDTH-1:0] out_payload      
 );
      
   always @* begin
     in_ready <= out_ready || ~out_valid;
   end
   
   always @ (negedge reset_n, posedge clk) begin
      if (!reset_n) begin
         out_valid <= 0;
         out_payload <= 0;
      end else begin
         if (in_valid) begin
           out_valid <= 1;
         end else if (out_ready) begin
           out_valid <= 0;
         end
         
         if(in_valid && in_ready) begin
            out_payload <= in_payload;
         end
      end
   end

endmodule
