//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module receives raw TLP packets in from the PCIe HIP core and it assigns a new channel valid to
//	the packet on egress.  The new channel value is used to demultiplex the packet stream based on the
//	contents of the TLP packet.  There are 4 possible demux paths which are encoded into the 2 least
//	significant bits of the egress packet channel feild:
//	
//		2'b00 = a request TLP which should be destined for a BAR master
//		2'b01 = a completion TLP which should be destined the appropriate tag manager
//		2'b10 = a message TLP which should be destined for an RX message FIFO
//		2'b11 = a bad decode occured.  You can do whatever you want with this data.
//		
//	The most significant 6 bits of the channel field are encoded based on the type of TLP decode that
//	was detected:
//	
//		request TLP = the encoded BARDEC value that was sent in with the packet, note this is not valid until the cycle after SOP
//		completion TLP = the least significant 6 bits of the tag field
//		message TLP = zero
//		bad decode = the raw 6 lsb bits of BARDEC
//	
	
module pcie_tlp_classify_64b (

	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// TLP Input Interface - PCIe ST
	input  wire	[ 63 : 0 ]	asi_snk_data,
	output wire				asi_snk_ready,
	input  wire				asi_snk_valid,
	input  wire				asi_snk_startofpacket,
	input  wire				asi_snk_endofpacket,
	input  wire	[  7 : 0 ]	asi_snk_channel,
	
	// TLP Output Interface - PCIe ST
	output reg	[ 63 : 0 ]	aso_src_data,
	input  wire				aso_src_ready,
	output wire				aso_src_valid,
	output reg				aso_src_startofpacket,
	output reg				aso_src_endofpacket,
	output reg	[  7 : 0 ]	aso_src_channel

);

wire [  4 : 0 ] tlp_type_field;
wire [  7 : 0 ] tlp_tag_field;
wire [  7 : 0 ]	new_channel_value;
wire [  7 : 0 ]	calc_next_channel;
reg	 [  7 : 0 ]	hold_next_channel;

reg  [ 63 : 0 ] pipe_data;
reg  			pipe_startofpacket;
reg 			pipe_endofpacket;
wire			pipe_ready;
wire			pipe_valid;

reg  [ 65 : 0 ] in_payload_pipe;
wire [ 65 : 0 ] out_payload_pipe;
reg  [ 73 : 0 ] in_payload;
wire [ 73 : 0 ] out_payload;

assign tlp_type_field = pipe_data[ 28 : 24 ];
assign tlp_tag_field  = asi_snk_data[ 15 : 8 ];

assign calc_next_channel =	( tlp_type_field[ 4 : 3 ] == 2'b10 ) ? ( { 6'h0, 2'b10 } ) :				// message TLP mux out channel 2'b10
							( tlp_type_field == 5'b01010 ) ? ( { tlp_tag_field[ 5 : 0 ], 2'b01 } ) :	// completion TLP mux out channel 2'b01
							( asi_snk_channel == 8'h01 ) ? ( { 6'h0, 2'b00 } ) :						// request TLP mux out channel 2'b00
							( asi_snk_channel == 8'h02 ) ? ( { 6'h1, 2'b00 } ) :						// request TLP mux out channel 2'b00
							( asi_snk_channel == 8'h04 ) ? ( { 6'h2, 2'b00 } ) :						// request TLP mux out channel 2'b00
							( asi_snk_channel == 8'h08 ) ? ( { 6'h3, 2'b00 } ) :						// request TLP mux out channel 2'b00
							( asi_snk_channel == 8'h10 ) ? ( { 6'h4, 2'b00 } ) :						// request TLP mux out channel 2'b00
							( asi_snk_channel == 8'h20 ) ? ( { 6'h5, 2'b00 } ) :						// request TLP mux out channel 2'b00
							( { asi_snk_channel[ 5 : 0 ], 2'b11 } );									// invalid decode mux out channel 2'b11

assign new_channel_value = ( pipe_startofpacket ) ? ( calc_next_channel ) : ( hold_next_channel );

always @ ( posedge clk or posedge reset ) begin

	if( reset ) begin
	
		hold_next_channel <= 0;
		
	end else begin
	
		if( pipe_startofpacket ) begin
	
			hold_next_channel <= calc_next_channel;
			
		end
	end
end

//
// input Pipeline
//
pcie_tlp_classify_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 66 ) ) inpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (asi_snk_ready),
    .in_valid       (asi_snk_valid), 
    .in_payload     (in_payload_pipe),
    .out_ready      ( ( pipe_startofpacket ) ? ( pipe_ready & pipe_valid & asi_snk_valid ) : ( pipe_ready ) ), 
    .out_valid      (pipe_valid), 
    .out_payload    (out_payload_pipe)
);

//
// output Pipeline
//
pcie_tlp_classify_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 74 ) ) outpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (pipe_ready),
    .in_valid       (pipe_valid), 
    .in_payload     (in_payload),
    .out_ready      (aso_src_ready), 
    .out_valid      (aso_src_valid), 
    .out_payload    (out_payload)
);

//
// Output Mapping
//
always @* begin
    in_payload_pipe <= { asi_snk_data, asi_snk_startofpacket, asi_snk_endofpacket};
    { pipe_data, pipe_startofpacket, pipe_endofpacket } <= out_payload_pipe;

    in_payload <= { pipe_data, pipe_startofpacket, pipe_endofpacket, new_channel_value};
    { aso_src_data, aso_src_startofpacket, aso_src_endofpacket, aso_src_channel } <= out_payload;
end

endmodule

//  --------------------------------------------------------------------------------
// | single buffered pipeline stage
//  --------------------------------------------------------------------------------
module pcie_tlp_classify_64b_1stage_pipeline  
#( parameter PAYLOAD_WIDTH = 8 )
 ( input                          clk,
   input                          reset_n, 
   output reg                     in_ready,
   input                          in_valid,   
   input      [PAYLOAD_WIDTH-1:0] in_payload,
   input                          out_ready,   
   output reg                     out_valid,
   output reg [PAYLOAD_WIDTH-1:0] out_payload      
 );
      
   always @* begin
     in_ready <= out_ready || ~out_valid;
   end
   
   always @ (negedge reset_n, posedge clk) begin
      if (!reset_n) begin
         out_valid <= 0;
         out_payload <= 0;
      end else begin
         if (in_valid) begin
           out_valid <= 1;
         end else if (out_ready) begin
           out_valid <= 0;
         end
         
         if(in_valid && in_ready) begin
            out_payload <= in_payload;
         end
      end
   end

endmodule
