// synthesis translate_off
`timescale 1ns / 1ps
// synthesis translate_on

module simulationPrint
(

	// Clock I/F
	input										csi_clock_clk,
	input										csi_clock_reset_n,
	
	// Write Avalon Slave I/F
	input	wire	[ 3 : 0 ]					avs_write_slave_address,
	input	wire	[ 3 : 0 ]					avs_write_slave_byteenable,
	input	wire								avs_write_slave_write,
	output	wire								avs_write_slave_waitrequest,
	input	wire	[ 31 : 0 ]					avs_write_slave_writedata

);

//
// These are the address decodes for accesses to this peripheral.  A write access to these address offsets
// will perform the associated function.
//
localparam	TIME_PRINT				= 4'h0;		// prints a time stamp message immediately
localparam	STOP_SUCCESS			= 4'h1;		// prints a success message and stops the simulation
localparam	STOP_FAIL				= 4'h2;		// prints a failure message and stops the simulation
localparam	ENABLE_MODULE_PRINT		= 4'h3;		// when enabled this module name hierarchy will be printed before each print function executes
localparam	DISABLE_MODULE_PRINT	= 4'h4;
localparam	PRINT_HEX_RAW			= 4'h5;		// prints the data written to this address as hex numbers
localparam	PRINT_HEX_TYPED			= 4'h6;		// prints the data written to this address as hex numbers followed by ".h"
localparam	PRINT_DEC_RAW			= 4'h7;		// prints the data written to this address as decimal numbers
localparam	PRINT_DEC_TYPED			= 4'h8;		// prints the data written to this address as decimal numbers followed by ".d"
localparam	PRINT_CHAR_RAW			= 4'h9;		// prints the data written to this address as ASCII characters
localparam	PRINT_LINE_FEED			= 4'hA;		// prints a line feed character
localparam	ENABLE_PRINT_STORE		= 4'hB;		// enable a print queue so that the PRINT functions will print into a queue instead of the console
localparam	DISABLE_PRINT_STORE		= 4'hC;		// disable the print queue storage
localparam	DUMP_PRINT_STORE		= 4'hD;		// dump the stored operations in the stored print queue

localparam PRINTSTORE_WIDTH =	40;	// 32 + 4 + 4


reg		moduleNamePrinting;
reg		printStore;

// synthesis translate_off
reg [ ( PRINTSTORE_WIDTH - 1 ) : 0 ] printStoreQueue[$];
reg	[ 3 : 0 ]					store_address;
reg	[ 3 : 0 ]					store_byteenable;
reg	[ 31 : 0 ]					store_writedata;
// synthesis translate_on

assign avs_write_slave_waitrequest	= 1'b0;

always @ (posedge csi_clock_clk or negedge csi_clock_reset_n) begin
	
	if(~csi_clock_reset_n) begin
	
		moduleNamePrinting	<= 0;
		printStore			<= 0;
	
	end else if(avs_write_slave_write) begin
	
// synthesis translate_off

		case(avs_write_slave_address)
			
			ENABLE_MODULE_PRINT	:	begin
										moduleNamePrinting <= 1'b1;
										$write("%m\nEnabling Module Name Printing...\n");
									end

			DISABLE_MODULE_PRINT	:	begin
											moduleNamePrinting <= 1'b0;
											$write("%m\nDisabling Module Name Printing...\n");
										end

			TIME_PRINT	:	begin
								if(moduleNamePrinting) 
									$write("%m\n");
								$write("simTime: %0d ns\n", $time);
							end
			
			STOP_SUCCESS	:	begin
									if(moduleNamePrinting) 
										$write("%m\n");
									$write("SUCCESS STOP - Stopping Simulation at simTime: %0d ns\n", $time);
									$stop;
								end
							
			STOP_FAIL	:	begin
									if(moduleNamePrinting) 
										$write("%m\n");
									$write("FAILURE STOP - Stopping Simulation at simTime: %0d ns\n", $time);
									$stop;
								end
							
			PRINT_HEX_RAW	:	begin
									if( printStore ) begin
									
										printStoreQueue.push_back( { avs_write_slave_address, avs_write_slave_byteenable, avs_write_slave_writedata } );
										
									end else begin
										if(moduleNamePrinting) 
											$write("%m\n");
										case(avs_write_slave_byteenable)
										
											4'b0001	: $write("%2H", avs_write_slave_writedata[  7 : 0 ]);
											4'b0011	: $write("%4H", avs_write_slave_writedata[ 15 : 0 ]);
											4'b0111	: $write("%6H", avs_write_slave_writedata[ 23 : 0 ]);
											4'b1111	: $write("%8H", avs_write_slave_writedata[ 31 : 0 ]);
											default	: $write("Invalid byte enable alignment during PRINT_HEX_RAW\n");
											
										endcase
									end
								end
							
			PRINT_HEX_TYPED	:	begin
									if( printStore ) begin
									
										printStoreQueue.push_back( { avs_write_slave_address, avs_write_slave_byteenable, avs_write_slave_writedata } );
										
									end else begin
										if(moduleNamePrinting) 
											$write("%m\n");
										case(avs_write_slave_byteenable)
										
											4'b0001	: $write("%2H.h", avs_write_slave_writedata[  7 : 0 ]);
											4'b0011	: $write("%4H.h", avs_write_slave_writedata[ 15 : 0 ]);
											4'b0111	: $write("%6H.h", avs_write_slave_writedata[ 23 : 0 ]);
											4'b1111	: $write("%8H.h", avs_write_slave_writedata[ 31 : 0 ]);
											default	: $write("Invalid byte enable alignment during PRINT_HEX_TYPED\n");
											
										endcase
									end
								end
							
			PRINT_DEC_RAW	:	begin
									if( printStore ) begin
									
										printStoreQueue.push_back( { avs_write_slave_address, avs_write_slave_byteenable, avs_write_slave_writedata } );
										
									end else begin
										if(moduleNamePrinting) 
											$write("%m\n");
										case(avs_write_slave_byteenable)
										
											4'b0001	: $write("%0d", avs_write_slave_writedata[  7 : 0 ]);
											4'b0011	: $write("%0d", avs_write_slave_writedata[ 15 : 0 ]);
											4'b0111	: $write("%0d", avs_write_slave_writedata[ 23 : 0 ]);
											4'b1111	: $write("%0d", avs_write_slave_writedata[ 31 : 0 ]);
											default	: $write("Invalid byte enable alignment during PRINT_DEC_RAW\n");
											
										endcase
									end
								end
							
			PRINT_DEC_TYPED	:	begin
									if( printStore ) begin
									
										printStoreQueue.push_back( { avs_write_slave_address, avs_write_slave_byteenable, avs_write_slave_writedata } );
										
									end else begin
										if(moduleNamePrinting) 
											$write("%m\n");
										case(avs_write_slave_byteenable)
										
											4'b0001	: $write("%0d.d", avs_write_slave_writedata[  7 : 0 ]);
											4'b0011	: $write("%0d.d", avs_write_slave_writedata[ 15 : 0 ]);
											4'b0111	: $write("%0d.d", avs_write_slave_writedata[ 23 : 0 ]);
											4'b1111	: $write("%0d.d", avs_write_slave_writedata[ 31 : 0 ]);
											default	: $write("Invalid byte enable alignment during PRINT_DEC_TYPED\n");
											
										endcase
									end
								end
							
			PRINT_CHAR_RAW	:	begin
									if( printStore ) begin
									
										printStoreQueue.push_back( { avs_write_slave_address, avs_write_slave_byteenable, avs_write_slave_writedata } );
										
									end else begin
										if(moduleNamePrinting) 
											$write("%m\n");
										case(avs_write_slave_byteenable)
										
											4'b0001	: $write("%c", avs_write_slave_writedata[  7 : 0 ]);
											4'b0011	: $write("%c%c", avs_write_slave_writedata[  7 : 0 ], avs_write_slave_writedata[ 15 : 8 ]);
											4'b0111	: $write("%c%c%c", avs_write_slave_writedata[  7 : 0 ], avs_write_slave_writedata[ 15 : 8 ], avs_write_slave_writedata[ 23 : 16 ]);
											4'b1111	: $write("%c%c%c%c", avs_write_slave_writedata[  7 : 0 ], avs_write_slave_writedata[ 15 : 8 ], avs_write_slave_writedata[ 23 : 16 ], avs_write_slave_writedata[ 31 : 24 ]);
											default	: $write("Invalid byte enable alignment during PRINT_CHAR_RAW\n");
											
										endcase
									end
								end
							
			PRINT_LINE_FEED	:	begin
									if( printStore ) begin
									
										printStoreQueue.push_back( { avs_write_slave_address, avs_write_slave_byteenable, avs_write_slave_writedata } );
										
									end else begin
										if(moduleNamePrinting) 
											$write("%m\n");
										$write("\n");
									end
								end
							
			ENABLE_PRINT_STORE :	begin
										printStore <= 1'b1;
										$write("%m\nEnabling Print Storage...\n");
									end
									
			DISABLE_PRINT_STORE :	begin
										printStore <= 1'b0;
										$write("%m\nDisabling Print Storage...\n");
									end
									
			DUMP_PRINT_STORE :	begin
									if(moduleNamePrinting) 
										$write("%m\n");
									while( printStoreQueue.size() > 0 ) begin
								
										{ store_address, store_byteenable, store_writedata } = printStoreQueue.pop_front();

										case(store_address)

											PRINT_HEX_RAW	:	begin
																	case(store_byteenable)
																	
																		4'b0001	: $write("%2H", store_writedata[  7 : 0 ]);
																		4'b0011	: $write("%4H", store_writedata[ 15 : 0 ]);
																		4'b0111	: $write("%6H", store_writedata[ 23 : 0 ]);
																		4'b1111	: $write("%8H", store_writedata[ 31 : 0 ]);
																		default	: $write("Invalid byte enable alignment during PRINT_HEX_RAW\n");
																		
																	endcase
																end
															
											PRINT_HEX_TYPED	:	begin
																	case(store_byteenable)
																	
																		4'b0001	: $write("%2H.h", store_writedata[  7 : 0 ]);
																		4'b0011	: $write("%4H.h", store_writedata[ 15 : 0 ]);
																		4'b0111	: $write("%6H.h", store_writedata[ 23 : 0 ]);
																		4'b1111	: $write("%8H.h", store_writedata[ 31 : 0 ]);
																		default	: $write("Invalid byte enable alignment during PRINT_HEX_TYPED\n");
																		
																	endcase
																end
															
											PRINT_DEC_RAW	:	begin
																	case(store_byteenable)
																	
																		4'b0001	: $write("%0d", store_writedata[  7 : 0 ]);
																		4'b0011	: $write("%0d", store_writedata[ 15 : 0 ]);
																		4'b0111	: $write("%0d", store_writedata[ 23 : 0 ]);
																		4'b1111	: $write("%0d", store_writedata[ 31 : 0 ]);
																		default	: $write("Invalid byte enable alignment during PRINT_DEC_RAW\n");
																		
																	endcase
																end
															
											PRINT_DEC_TYPED	:	begin
																	case(store_byteenable)
																	
																		4'b0001	: $write("%0d.d", store_writedata[  7 : 0 ]);
																		4'b0011	: $write("%0d.d", store_writedata[ 15 : 0 ]);
																		4'b0111	: $write("%0d.d", store_writedata[ 23 : 0 ]);
																		4'b1111	: $write("%0d.d", store_writedata[ 31 : 0 ]);
																		default	: $write("Invalid byte enable alignment during PRINT_DEC_TYPED\n");
																		
																	endcase
																end
															
											PRINT_CHAR_RAW	:	begin
																	case(store_byteenable)
																	
																		4'b0001	: $write("%c", store_writedata[  7 : 0 ]);
																		4'b0011	: $write("%c%c", store_writedata[  7 : 0 ], store_writedata[ 15 : 8 ]);
																		4'b0111	: $write("%c%c%c", store_writedata[  7 : 0 ], store_writedata[ 15 : 8 ], store_writedata[ 23 : 16 ]);
																		4'b1111	: $write("%c%c%c%c", store_writedata[  7 : 0 ], store_writedata[ 15 : 8 ], store_writedata[ 23 : 16 ], store_writedata[ 31 : 24 ]);
																		default	: $write("Invalid byte enable alignment during PRINT_CHAR_RAW\n");
																		
																	endcase
																end
															
											PRINT_LINE_FEED	:	begin
																		$write("\n");
																end
															
											default	:	begin
															if(moduleNamePrinting) 
																$write("\n%m\n");
															$write("simulationPrint: in storePrint Dump bad address of 0x%H\n", store_address);
														end
										endcase
									end
								end
								
			default	:	begin
							if(moduleNamePrinting) 
								$write("%m\n");
							$write("simulationPrint: bad address of 0x%H\n", avs_write_slave_address);
						end
		endcase	
	
// synthesis translate_on
	end
	
end

endmodule
