//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module translates Avalon Slave transactions into PCIe request TLPs.
//	

module slave2tlp_64b
#(
	parameter SLAVE_ADDRESS_WIDTH	= 9,
	parameter MAX_READ_REQ_SIZE		= 5,
	parameter AVAILABLE_TAG_COUNT	= 16,
	parameter BASE_TAG_VALUE		= 0
)
(
	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// DEVCSR Input Interface - Avalon ST
	input  wire [ 15 : 0 ]	asi_devcsrsnk_data,
	
	// BUSDEV Input Interface - Avalon ST
	input  wire [ 12 : 0 ]	asi_busdevsnk_data,
	
	// TLP Slave Interface - Avalon MM
	input  wire	[ ( SLAVE_ADDRESS_WIDTH - 1 ) : 0 ]	avs_s0_address,
	input  wire	[  7 : 0 ]	avs_s0_byteenable,
	input  wire	[  9 : 0 ]	avs_s0_burstcount,
	input  wire				avs_s0_read,
	input  wire				avs_s0_write,
	input  wire	[ 63 : 0 ]	avs_s0_writedata,
	output wire				avs_s0_waitrequest,
	output wire	[ 63 : 0 ]	avs_s0_readdata,
	output wire				avs_s0_readdatavalid,

	// CSR Slave Interface - Avalon MM
	input  wire	[  1 : 0 ]	avs_csr_address,
	input  wire	[  3 : 0 ]	avs_csr_byteenable,
	input  wire				avs_csr_read,
	input  wire				avs_csr_write,
	input  wire	[ 31 : 0 ]	avs_csr_writedata,
	output wire	[ 31 : 0 ]	avs_csr_readdata,
	
	// TAG Request Master Output Interface - Avalon MM
	output wire	[  7 : 0 ]	avm_tagreq_address,		// 64 tags maximum
	output wire	[  3 : 0 ]	avm_tagreq_byteenable,
	output wire				avm_tagreq_write,
	output wire	[ 31 : 0 ]	avm_tagreq_writedata,
	input  wire				avm_tagreq_waitrequest,

	// Avalon Slave Read Data Input Interface - Avalon ST
	input  wire	[ 63 : 0 ]	asi_rddatasnk_data,
	output wire				asi_rddatasnk_ready,
	input  wire				asi_rddatasnk_valid,
	input  wire				asi_rddatasnk_startofpacket,
	input  wire				asi_rddatasnk_endofpacket,
	
	// EOP LBE Output Interface - Avalon ST
	output wire	[  4 : 0 ]	aso_eoplbesrc_data,
	input  wire				aso_eoplbesrc_ready,
	output wire				aso_eoplbesrc_valid,

	// TLP Output Interface - PCIe ST
	output wire	[ 63 : 0 ]	aso_tlpsrc_data,
	input  wire				aso_tlpsrc_ready,
	output wire				aso_tlpsrc_valid,
	output wire				aso_tlpsrc_startofpacket,
	output wire				aso_tlpsrc_endofpacket
	
);

localparam TAG_COUNTER_WIDTH = log2ceil( AVAILABLE_TAG_COUNT );

//
// state definitions
//
localparam INITIAL_HEADER_1_0_STATE		= 3'h0;
localparam INITIAL_HEADER_3_2_STATE 	= 3'h1;
localparam WRITE_DATA_STATE 			= 3'h2;
localparam CONTINUE_HEADER_1_0_STATE	= 3'h3;
localparam CONTINUE_HEADER_3_2_STATE	= 3'h4;
localparam FINAL_HEADER_3_2_STATE		= 3'h5;
localparam FINAL_WRITE_DATA_STATE		= 3'h6;

reg  [  2 : 0 ]	state;

wire			waitrequest_inv;

wire			in_pipe_1_in_ready;
wire			in_pipe_1_out_valid;
wire [ ( SLAVE_ADDRESS_WIDTH - 1 ) : 0 ]	p1_avs_s0_address;
wire [  7 : 0 ]	p1_avs_s0_byteenable;
wire [  9 : 0 ]	p1_avs_s0_burstcount;
wire			p1_avs_s0_read;
wire			p1_avs_s0_write;
wire [ 63 : 0 ]	p1_avs_s0_writedata;

//
// CSR access
//
reg  [ 31 : 0 ]	tlp_lower_address_bits;
reg  [ 31 : 0 ]	tlp_upper_address_bits;
wire [ 63 : 0 ] tlp_64bit_address_bits;
reg  [  2 : 0 ] tlp_tc_bits;

//
// Misc Status Inputs
//
wire [  2 : 0 ]	max_payload_size_field;
wire [ 12 : 0 ] max_payload_size;

wire [  2 : 0 ]	max_read_request_size_field;
wire [ 12 : 0 ] max_read_request_size;

wire 			extended_tag_field_enable_field;

wire [  7 : 0 ] bus_number_field;

wire [  4 : 0 ] dev_number_field;

//
// readdata output pipeline
//
wire			readdata_outpipe_valid;

//
// Request output Pipeline
//
wire			request_outpipe_ready;
wire			request_outpipe_valid;
wire [ 63 : 0 ]	request_data;
wire			request_sop;
wire			request_eop;

//
// EOP LBE output Pipeline
//
wire			eoplbe_outpipe_ready;
wire			eoplbe_outpipe_valid;
wire			write_req_sub_1dw;
wire [  3 : 0 ]	write_req_lbe_bits;

//
// Write burst count monitor
//
wire [  9 : 0 ]	current_burst_count_plus1;
reg  [  9 : 0 ]	current_burst_count;
reg  [  9 : 0 ]	reg_this_write_request_burst_count;

//
// Misc Computed Values
//
wire [ 63 : 0 ] merged_64bit_slave_address;
wire [ 63 : 0 ] pipe_merged_64bit_slave_address;
wire [ ( SLAVE_ADDRESS_WIDTH - 1 ) : 0 ]	current_slave_address;
wire [ 63 : 0 ] merged_64bit_current_slave_address;

wire [  2 : 0 ]	dw_0_be_sum;
wire [  2 : 0 ]	dw_1_be_sum;

wire			first_dw_is_unaligned;
wire			last_dw_is_empty;

wire [ 11 : 0 ] read_base_start_address_mask;
wire [ 11 : 0 ] write_base_start_address_mask;

wire [ 11 : 0 ] read_base_start_address;
wire [ 11 : 0 ] write_base_start_address;

wire [ 12 : 0 ] allowed_read_bytes_in_first_request;
wire [ 12 : 0 ] allowed_write_bytes_in_first_request;

wire [  9 : 0 ] allowed_read_burstcount_in_first_request;
wire [  9 : 0 ] allowed_write_burstcount_in_first_request;

wire [  9 : 0 ] p1_allowed_read_burstcount_in_first_request;
wire [  9 : 0 ] p1_allowed_write_burstcount_in_first_request;

wire [  9 : 0 ] actual_read_burstcount_in_first_request;
wire [  9 : 0 ] actual_write_burstcount_in_first_request;

wire [  9 : 0 ] actual_read_dw_in_first_request;
wire [  9 : 0 ] actual_write_dw_in_first_request;

wire [ 12 : 0 ] sum_4k_limit;
wire [ 12 : 0 ] difference_4k_limit;

wire			more_than_one_read_request;
wire			more_than_one_write_request;
wire			p1_more_than_one_read_request;
wire			p1_more_than_one_write_request;

wire			only_one_tlp_required;

wire			another_full_read_burst_required;
wire			another_full_write_burst_required;

wire [  8 : 0 ] read_burstcount_in_this_request;
wire [  8 : 0 ] write_burstcount_in_this_request;

wire [  9 : 0 ] read_dw_in_this_request;
wire [  9 : 0 ] write_dw_in_this_request;
reg  [  9 : 0 ] reg_write_dw_in_this_request;

wire			use_4dw_header;

//
// TAG management
//
reg  [ ( TAG_COUNTER_WIDTH - 1 ) : 0 ]	current_tag_value;
reg  [ ( TAG_COUNTER_WIDTH - 1 ) : 0 ]	next_tag_value;
wire [ ( TAG_COUNTER_WIDTH - 1 ) : 0 ]	next_tag_value_plus1;
wire			advance_next_tag;
wire			next_tag_available;
wire			advance_current_tag;
reg 			tag_allocate_error;
reg	    		tag_overrun_error;

//
// TLP header word 0
//
wire [  0 : 0 ] tlp_res3_field;
wire [  1 : 0 ] tlp_fmt_field;
wire [  4 : 0 ] tlp_type_field;
wire [  0 : 0 ] tlp_res2_field;
wire [  2 : 0 ] tlp_tc_field;
wire [  3 : 0 ] tlp_res1_field;
wire [  0 : 0 ] tlp_td_field;
wire [  0 : 0 ] tlp_ep_field;
wire [  1 : 0 ] tlp_attr_field;
wire [  1 : 0 ] tlp_res0_field;
wire [  9 : 0 ] tlp_length_field;

//
// TLP header word 1
//
wire [ 15 : 0 ] tlp_reqid_field;
wire [  7 : 0 ] tlp_tag_field;
wire [  3 : 0 ] tlp_lbe_field;
wire [  3 : 0 ] tlp_fbe_field;

//
// TLP header word 2 and 3
//
wire [ 61 : 0 ] tlp_address_field;
wire [  1 : 0 ] tlp_res4_field;

wire			tlp_a2_value;

wire [ 31 : 0 ] request_header_word_0;
wire [ 31 : 0 ] request_header_word_1;
wire [ 31 : 0 ] request_header_word_2_3dw;
wire [ 31 : 0 ] request_header_word_2_4dw;
wire [ 31 : 0 ] request_header_word_3_4dw;

assign request_header_word_0 = {
			tlp_res3_field,
			tlp_fmt_field,
			tlp_type_field,
			tlp_res2_field,
			tlp_tc_field,
			tlp_res1_field,
			tlp_td_field,
			tlp_ep_field,
			tlp_attr_field,
			tlp_res0_field,
			tlp_length_field
		};

assign request_header_word_1 = {
		tlp_reqid_field,
		tlp_tag_field,
		tlp_lbe_field,
		tlp_fbe_field
		};

assign request_header_word_2_3dw = {
			tlp_address_field[ 29 : 0 ],
			tlp_res4_field
		};

assign request_header_word_2_4dw = {
			tlp_address_field[ 61 : 30 ]
		};

assign request_header_word_3_4dw = {
			tlp_address_field[ 29 : 0 ],
			tlp_res4_field
		};

assign tlp_res3_field =  1'b0;

assign tlp_fmt_field =  { ~p1_avs_s0_read , use_4dw_header };

assign tlp_type_field =  5'b00000;

assign tlp_res2_field =  1'b0;

assign tlp_tc_field =  tlp_tc_bits;

assign tlp_res1_field =  4'b0000;

assign tlp_td_field =  1'b0;

assign tlp_ep_field =  1'b0;

assign tlp_attr_field =  2'b00;

assign tlp_res0_field =  2'b00;

assign tlp_length_field =	( state == INITIAL_HEADER_1_0_STATE ) ?
								( p1_avs_s0_read ) ? 
									( actual_read_dw_in_first_request ) : 
									( actual_write_dw_in_first_request ) :
								( p1_avs_s0_read ) ? 
									( read_dw_in_this_request ) : 
									( write_dw_in_this_request );

assign tlp_reqid_field =  { bus_number_field, dev_number_field, 3'b000 };

assign tlp_tag_field =  BASE_TAG_VALUE + next_tag_value;

assign tlp_lbe_field =  ( state == INITIAL_HEADER_1_0_STATE ) ? 
							( first_dw_is_unaligned ) ? 
								( 4'h0 ) : 
								( p1_avs_s0_byteenable[ 7 : 4 ] ) : 
							( 4'hF );

assign tlp_fbe_field =  ( state == INITIAL_HEADER_1_0_STATE ) ? 
							( first_dw_is_unaligned ) ? 
								( p1_avs_s0_byteenable[ 7 : 4 ] ) : 
								( p1_avs_s0_byteenable[ 3 : 0 ] ) : 
							( p1_avs_s0_write && ( write_dw_in_this_request == 2 ) && write_req_sub_1dw ) ? 
								( p1_avs_s0_byteenable[ 3 : 0 ] ) : 
								( 4'hF );

assign tlp_address_field = ( state == INITIAL_HEADER_3_2_STATE ) ? ( { pipe_merged_64bit_slave_address[ 63 : 3 ], tlp_a2_value } ) : ( merged_64bit_current_slave_address[ 63 : 2 ] );

assign tlp_res4_field = 2'b00;

assign tlp_a2_value =  first_dw_is_unaligned;

//
// Misc Computed Values
//
assign merged_64bit_slave_address = { tlp_64bit_address_bits[ 63 : ( SLAVE_ADDRESS_WIDTH + 3 ) ], avs_s0_address, 3'b000 };

assign pipe_merged_64bit_slave_address = { tlp_64bit_address_bits[ 63 : ( SLAVE_ADDRESS_WIDTH + 3 ) ], p1_avs_s0_address, 3'b000 };

assign current_slave_address = p1_avs_s0_address + current_burst_count;

assign merged_64bit_current_slave_address = { tlp_64bit_address_bits[ 63 : ( SLAVE_ADDRESS_WIDTH + 3 ) ], current_slave_address, 3'b000 };

assign dw_0_be_sum = p1_avs_s0_byteenable[0] + p1_avs_s0_byteenable[1] + p1_avs_s0_byteenable[2] + p1_avs_s0_byteenable[3];
assign dw_1_be_sum = p1_avs_s0_byteenable[4] + p1_avs_s0_byteenable[5] + p1_avs_s0_byteenable[6] + p1_avs_s0_byteenable[7];

assign first_dw_is_unaligned = ( ( dw_1_be_sum > 0 ) && ( dw_0_be_sum == 0 ) ) ? ( 1'b1 ) : ( 1'b0 );
assign last_dw_is_empty = ( dw_1_be_sum == 0 ) ? ( 1'b1 ) : ( 1'b0 );

assign read_base_start_address_mask =	( max_read_request_size_field == 1 ) ? ( 12'h0FF ) :	//  256 - 0x100  - 0x0FF
										( max_read_request_size_field == 2 ) ? ( 12'h1FF ) :	//  512 - 0x200  - 0x1FF
										( max_read_request_size_field == 3 ) ? ( 12'h3FF ) :	// 1024 - 0x400  - 0x3FF
										( max_read_request_size_field == 4 ) ? ( 12'h7FF ) :	// 2048 - 0x800  - 0x7FF
										( max_read_request_size_field == 5 ) ? ( 12'hFFF ) :	// 4096 - 0x1000 - 0xFFF
										( 12'h07F );											//  128 - 0x80   - 0x07F
								
assign write_base_start_address_mask =	( max_payload_size_field == 1 ) ? ( 12'h0FF ) :	//  256 - 0x100  - 0x0FF
										( max_payload_size_field == 2 ) ? ( 12'h1FF ) :	//  512 - 0x200  - 0x1FF
										( max_payload_size_field == 3 ) ? ( 12'h3FF ) :	// 1024 - 0x400  - 0x3FF
										( max_payload_size_field == 4 ) ? ( 12'h7FF ) :	// 2048 - 0x800  - 0x7FF
										( max_payload_size_field == 5 ) ? ( 12'hFFF ) :	// 4096 - 0x1000 - 0xFFF
										( 12'h07F );									//  128 - 0x80   - 0x07F
								
assign read_base_start_address = merged_64bit_slave_address[ 11 : 0 ] & read_base_start_address_mask;

assign write_base_start_address = merged_64bit_slave_address[ 11 : 0 ] & write_base_start_address_mask;

assign sum_4k_limit = merged_64bit_slave_address[ 11 : 0 ] + { avs_s0_burstcount, 3'h0 };

assign difference_4k_limit = { avs_s0_burstcount, 3'h0 } - sum_4k_limit[ 11 : 0 ];

assign allowed_read_bytes_in_first_request =	( ( sum_4k_limit[12] == 1'b1 ) && ( max_read_request_size > difference_4k_limit ) ) ? 
													( difference_4k_limit ) :
													( max_read_request_size >= { avs_s0_burstcount, 3'h0 } ) ?
														( { avs_s0_burstcount, 3'h0 } ) :
														( max_read_request_size - { 1'b0, read_base_start_address } );

assign allowed_write_bytes_in_first_request =	( ( sum_4k_limit[12] == 1'b1 ) && ( max_payload_size > difference_4k_limit ) ) ? 
													( difference_4k_limit ) :
													( max_payload_size >= { avs_s0_burstcount, 3'h0 } ) ?
														( { avs_s0_burstcount, 3'h0 } ) :
														( max_payload_size - { 1'b0, write_base_start_address } );

assign allowed_read_burstcount_in_first_request = allowed_read_bytes_in_first_request[ 12 : 3 ];

assign allowed_write_burstcount_in_first_request = allowed_write_bytes_in_first_request[ 12 : 3 ];

assign more_than_one_read_request = ( avs_s0_burstcount > allowed_read_burstcount_in_first_request ) ? ( 1'b1 ) : ( 1'b0 );

assign more_than_one_write_request = ( avs_s0_burstcount > allowed_write_burstcount_in_first_request ) ? ( 1'b1 ) : ( 1'b0 );

assign actual_read_burstcount_in_first_request =	( p1_more_than_one_read_request ) ? 
														( p1_allowed_read_burstcount_in_first_request ) : 
														( p1_avs_s0_burstcount );

assign actual_write_burstcount_in_first_request =	( p1_more_than_one_write_request ) ? 
														( p1_allowed_write_burstcount_in_first_request ) : 
														( p1_avs_s0_burstcount );

assign actual_read_dw_in_first_request =	( actual_read_burstcount_in_first_request == 1) ?
												( first_dw_is_unaligned | last_dw_is_empty ) ?
													( 10'h1 ) :
													( 10'h2 ) :
												( first_dw_is_unaligned ) ? 
													( { actual_read_burstcount_in_first_request[ 8 : 0 ], 1'b0 } - 1 ) : 
													( { actual_read_burstcount_in_first_request[ 8 : 0 ], 1'b0 } );

assign actual_write_dw_in_first_request =	( first_dw_is_unaligned ) ? 
												( { actual_write_burstcount_in_first_request[ 8 : 0 ], 1'b0 } - 1 ) : 
												( { actual_write_burstcount_in_first_request[ 8 : 0 ], 1'b0 } );

assign only_one_tlp_required =	( p1_avs_s0_read & ~p1_more_than_one_read_request ) | 
								( p1_avs_s0_write & ~p1_more_than_one_write_request );
								
assign another_full_read_burst_required = ( ( p1_avs_s0_burstcount - current_burst_count ) > max_read_request_size[ 12 : 3 ] ) ? ( 1'b1 ) : ( 1'b0 );

assign another_full_write_burst_required = ( ( p1_avs_s0_burstcount - current_burst_count ) > max_payload_size[ 12 : 3 ] ) ? ( 1'b1 ) : ( 1'b0 );

assign read_burstcount_in_this_request = ( another_full_read_burst_required ) ? ( max_read_request_size[ 12 : 3 ] ) : ( ( p1_avs_s0_burstcount - current_burst_count ) );

assign write_burstcount_in_this_request = ( another_full_write_burst_required ) ? ( max_payload_size[ 12 : 3 ] ) : ( ( p1_avs_s0_burstcount - current_burst_count ) );

assign read_dw_in_this_request = { read_burstcount_in_this_request, 1'b0 };

assign write_dw_in_this_request = { write_burstcount_in_this_request, 1'b0 };

assign use_4dw_header = ( tlp_upper_address_bits != 0 );

assign current_burst_count_plus1 = current_burst_count + 1;

//
// CSR access
//
assign avs_csr_readdata =	( avs_csr_address == 2'h0 ) ? ( tlp_lower_address_bits ) :
							( avs_csr_address == 2'h1 ) ? ( tlp_upper_address_bits ) :
							( avs_csr_address == 2'h2 ) ? ( { { 29 { 1'b0 } }, tlp_tc_bits } ) :
														  ( { { 27 { 1'b0 } }, extended_tag_field_enable_field, tag_allocate_error, tag_overrun_error, next_tag_available, avm_tagreq_write } );
							
assign tlp_64bit_address_bits = { tlp_upper_address_bits, tlp_lower_address_bits };

always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin

		tlp_lower_address_bits <= 0;
		tlp_upper_address_bits <= 0;
		tlp_tc_bits <= 0;

	end else begin
		
		if( avs_csr_write ) begin
			case( avs_csr_address )
			2'h0: begin
				if( avs_csr_byteenable[0] )	tlp_lower_address_bits[  7 :  0 ] <= avs_csr_writedata[  7 :  0 ];
				if( avs_csr_byteenable[1] )	tlp_lower_address_bits[ 15 :  8 ] <= avs_csr_writedata[ 15 :  8 ];
				if( avs_csr_byteenable[2] )	tlp_lower_address_bits[ 23 : 16 ] <= avs_csr_writedata[ 23 : 16 ];
				if( avs_csr_byteenable[3] )	tlp_lower_address_bits[ 31 : 24 ] <= avs_csr_writedata[ 31 : 24 ];
			end
			2'h1: begin
				if( avs_csr_byteenable[0] )	tlp_upper_address_bits[  7 :  0 ] <= avs_csr_writedata[  7 :  0 ];
				if( avs_csr_byteenable[1] )	tlp_upper_address_bits[ 15 :  8 ] <= avs_csr_writedata[ 15 :  8 ];
				if( avs_csr_byteenable[2] )	tlp_upper_address_bits[ 23 : 16 ] <= avs_csr_writedata[ 23 : 16 ];
				if( avs_csr_byteenable[3] )	tlp_upper_address_bits[ 31 : 24 ] <= avs_csr_writedata[ 31 : 24 ];
			end
			2'h2: begin
				if( avs_csr_byteenable[0] )	tlp_tc_bits[  2 :  0 ] <= avs_csr_writedata[  2 :  0 ];
			end
			default: begin
			end
			endcase

		end
	end
end

//
// Misc Status Inputs
//
assign max_payload_size_field = asi_devcsrsnk_data[  7 : 5 ];
assign max_payload_size =	( max_payload_size_field == 1 ) ? ( 13'd256 ) :
							( max_payload_size_field == 2 ) ? ( 13'd512 ) :
							( max_payload_size_field == 3 ) ? ( 13'd1024 ) :
							( max_payload_size_field == 4 ) ? ( 13'd2048 ) :
							( max_payload_size_field == 5 ) ? ( 13'd4096 ) :
							( 13'd128 );

assign max_read_request_size_field = ( asi_devcsrsnk_data[ 14 : 12 ] > MAX_READ_REQ_SIZE ) ? ( MAX_READ_REQ_SIZE ) : ( asi_devcsrsnk_data[ 14 : 12 ] );
assign max_read_request_size =	( max_read_request_size_field == 1 ) ? ( 13'd256 ) :
								( max_read_request_size_field == 2 ) ? ( 13'd512 ) :
								( max_read_request_size_field == 3 ) ? ( 13'd1024 ) :
								( max_read_request_size_field == 4 ) ? ( 13'd2048 ) :
								( max_read_request_size_field == 5 ) ? ( 13'd4096 ) :
								( 13'd128 );
														
assign extended_tag_field_enable_field = asi_devcsrsnk_data[ 8 ];

assign bus_number_field = asi_busdevsnk_data[ 12 : 5 ];

assign dev_number_field = asi_busdevsnk_data[  4 : 0 ];

//
// Slave Interface control
//
assign avs_s0_waitrequest =	~in_pipe_1_in_ready;

assign waitrequest_inv =	( request_outpipe_ready & request_outpipe_valid ) ?
								( ( state == INITIAL_HEADER_3_2_STATE ) && p1_avs_s0_read && ~p1_more_than_one_read_request ) ? 
									( 1'b1 ) :
									( ( state == FINAL_HEADER_3_2_STATE ) && p1_avs_s0_read ) ? 
										( 1'b1 ) :
										( ( state == INITIAL_HEADER_3_2_STATE ) && p1_avs_s0_write && ~use_4dw_header && first_dw_is_unaligned ) ? 
											( 1'b1 ) :
											( ( state == WRITE_DATA_STATE ) || ( state == FINAL_WRITE_DATA_STATE ) ) ? 
												( 1'b1 ) :
												( 1'b0 ) :
								( 1'b0 );

//
// Request TLP control
//
assign request_outpipe_valid =	( ( ( in_pipe_1_out_valid & p1_avs_s0_read ) && ( next_tag_available || ( ( state != INITIAL_HEADER_1_0_STATE ) && ( state != CONTINUE_HEADER_1_0_STATE ) ) ) ) || ( in_pipe_1_out_valid & p1_avs_s0_write ) ) &&
								( request_outpipe_ready && eoplbe_outpipe_ready );

assign eoplbe_outpipe_valid = request_eop & request_outpipe_valid;

assign request_data =	( ( state == INITIAL_HEADER_1_0_STATE ) || ( state == CONTINUE_HEADER_1_0_STATE ) ) ? 
							( { request_header_word_1, request_header_word_0 } ) :
							( ( state == INITIAL_HEADER_3_2_STATE ) || ( state == CONTINUE_HEADER_3_2_STATE ) || ( state == FINAL_HEADER_3_2_STATE ) ) ? 
								( use_4dw_header ) ? 
									( { request_header_word_3_4dw, request_header_word_2_4dw } ) : 
									( { p1_avs_s0_writedata[ 63 : 32 ], request_header_word_2_3dw }) :
								( p1_avs_s0_writedata[ 63 : 0 ] );

assign request_sop = ( state == INITIAL_HEADER_1_0_STATE ) || ( state == CONTINUE_HEADER_1_0_STATE );

assign request_eop =	( p1_avs_s0_read ) ?
							( ( state == INITIAL_HEADER_3_2_STATE ) || ( state == CONTINUE_HEADER_3_2_STATE ) || ( state == FINAL_HEADER_3_2_STATE ) ) ? 
								( 1'b1 ) : 
								( 1'b0 ) :
							( ( state == INITIAL_HEADER_3_2_STATE ) && ( p1_avs_s0_burstcount == 1 ) && ( ~use_4dw_header ) && ( first_dw_is_unaligned ) ) ? 
								( 1'b1 ) :
								( ( state == INITIAL_HEADER_3_2_STATE ) && ( p1_avs_s0_burstcount != 1 ) && ( ~use_4dw_header ) && ( first_dw_is_unaligned ) && ( actual_write_burstcount_in_first_request == 1 ) ) ? 
									( 1'b1 ) :
									( ( ( state == WRITE_DATA_STATE ) || ( state == FINAL_WRITE_DATA_STATE ) ) && ( current_burst_count_plus1 >= reg_this_write_request_burst_count ) ) ? 
										( 1'b1 ) :
										( 1'b0 );

assign write_req_sub_1dw = last_dw_is_empty ;

assign write_req_lbe_bits = ( p1_avs_s0_burstcount == 1 ) ?
								( first_dw_is_unaligned ) ?
									( 4'h0 ) :
									( p1_avs_s0_byteenable[ 7 : 4 ] ) :
								( ( state == INITIAL_HEADER_3_2_STATE ) || ( ( state == WRITE_DATA_STATE ) && ( current_burst_count == 0 ) ) ) ?
									( actual_write_dw_in_first_request == 10'h01 ) ?
										( 4'h0 ) :
										( p1_avs_s0_byteenable[ 7 : 4 ] ) :
									( ( state == FINAL_WRITE_DATA_STATE ) && ( reg_write_dw_in_this_request == 2 ) && write_req_sub_1dw ) ?
										( 4'h0 ) :
										( last_dw_is_empty ) ?
											( p1_avs_s0_byteenable[ 3 : 0 ] ) :
											( p1_avs_s0_byteenable[ 7 : 4 ] );
																				

//
// Request TLP state sequence.
//
always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin

		state <= INITIAL_HEADER_1_0_STATE;
		current_burst_count <= 0;
		reg_this_write_request_burst_count <= 0;
		reg_write_dw_in_this_request <= 0;
		
	end else begin

		if( request_outpipe_ready & request_outpipe_valid ) begin
			case( state )
			INITIAL_HEADER_1_0_STATE: begin
			
				state <= INITIAL_HEADER_3_2_STATE;
				current_burst_count <= 0;
				
			end
			INITIAL_HEADER_3_2_STATE: begin
				
				if( p1_avs_s0_read ) begin								// read cycle
				
					if( p1_more_than_one_read_request ) begin				// we need multiple TLP requests
					
						state <= CONTINUE_HEADER_1_0_STATE;
						current_burst_count <= p1_allowed_read_burstcount_in_first_request;
						
					end else begin										// we only need ONE TLP request
					
						state <= INITIAL_HEADER_1_0_STATE;				// *** waitrequest deasserts here ***
						current_burst_count <= 0;
						
					end
				end else begin										// write cycle
				
					if( p1_avs_s0_burstcount == 1 ) begin					// single word burst
					
						if( use_4dw_header ) begin							// 4DW header

							state <= FINAL_WRITE_DATA_STATE;
							current_burst_count <= 0;
							reg_this_write_request_burst_count <= actual_write_burstcount_in_first_request;
							reg_write_dw_in_this_request <= write_dw_in_this_request;

						end else begin										// 3DW header
						
							if( first_dw_is_unaligned ) begin					// first dw is unaligned, write data is merged with header_2

								state <= INITIAL_HEADER_1_0_STATE;				// *** waitrequest deasserts here ***
								current_burst_count <= 0;

							end else begin										// first dw is aligned

								state <= FINAL_WRITE_DATA_STATE;
								current_burst_count <= 0;
								reg_this_write_request_burst_count <= actual_write_burstcount_in_first_request;
								reg_write_dw_in_this_request <= write_dw_in_this_request;

							end
						end
					end else begin										// multi word burst
					
						if( use_4dw_header ) begin							// 4DW header

							state <= ( p1_more_than_one_write_request ) ? ( WRITE_DATA_STATE ) : ( FINAL_WRITE_DATA_STATE );;
							current_burst_count <= 0;
							reg_this_write_request_burst_count <= actual_write_burstcount_in_first_request;
							reg_write_dw_in_this_request <= write_dw_in_this_request;

						end else begin										// 3DW header

							if( first_dw_is_unaligned ) begin					// first dw is unaligned, write data is merged with header_2

								if( actual_write_burstcount_in_first_request == 1 ) begin	// this will increment us to the next break boundry

									state <= CONTINUE_HEADER_1_0_STATE;				// *** waitrequest deasserts here ***
									current_burst_count <= current_burst_count + 1;

								end else begin										// this will NOT increment us to the next break boundry

									state <= ( p1_more_than_one_write_request ) ? ( WRITE_DATA_STATE ) : ( FINAL_WRITE_DATA_STATE );	// *** waitrequest deasserts here ***
									current_burst_count <= current_burst_count + 1;
									reg_this_write_request_burst_count <= actual_write_burstcount_in_first_request;
									reg_write_dw_in_this_request <= write_dw_in_this_request;
									
								end
							end else begin										// first dw is aligned

								state <= ( p1_more_than_one_write_request ) ? ( WRITE_DATA_STATE ) : ( FINAL_WRITE_DATA_STATE );
								current_burst_count <= 0;
								reg_this_write_request_burst_count <= actual_write_burstcount_in_first_request;
								reg_write_dw_in_this_request <= write_dw_in_this_request;

							end
						end
					end
				end
			end
			WRITE_DATA_STATE: begin
			
				if( current_burst_count_plus1 >= reg_this_write_request_burst_count ) begin 	// this is the last burst for this TLP
				
					state <= CONTINUE_HEADER_1_0_STATE;											// *** waitrequest deasserts here ***
					current_burst_count <= current_burst_count + 1;
					
				end else begin																	// this is NOT the last burst for this TLP
				
					current_burst_count <= current_burst_count + 1;								// *** waitrequest deasserts here ***
					
				end
			end
			CONTINUE_HEADER_1_0_STATE: begin
				
				if( p1_avs_s0_read ) begin		// read burst
				
					if( another_full_read_burst_required ) begin 	// we will create another TLP after this one
					
						state <= CONTINUE_HEADER_3_2_STATE;

					end else begin									// this is our final TLP

						state <= FINAL_HEADER_3_2_STATE;
					end
				end else begin				// write burst
				
					if( another_full_write_burst_required ) begin 	// we will create another TLP after this one
					
						state <= CONTINUE_HEADER_3_2_STATE;

					end else begin									// this is our final TLP

						state <= FINAL_HEADER_3_2_STATE;
					end
				end
			end
			CONTINUE_HEADER_3_2_STATE: begin
			
				if( p1_avs_s0_read ) begin						// read burst

					state <= CONTINUE_HEADER_1_0_STATE;
					current_burst_count <= current_burst_count + read_burstcount_in_this_request;
					
				end else begin								// write burst

					state <= WRITE_DATA_STATE;
					reg_this_write_request_burst_count <= current_burst_count + write_burstcount_in_this_request;
					reg_write_dw_in_this_request <= write_dw_in_this_request;

				end
			end
			FINAL_HEADER_3_2_STATE: begin
				if( p1_avs_s0_read ) begin						// read burst

					state <= INITIAL_HEADER_1_0_STATE;		// *** waitrequest deasserts here ***
					current_burst_count <= 0;
					
				end else begin								// write burst

					state <= FINAL_WRITE_DATA_STATE;
					reg_this_write_request_burst_count <= current_burst_count + write_burstcount_in_this_request;
					reg_write_dw_in_this_request <= write_dw_in_this_request;

				end
			end
			FINAL_WRITE_DATA_STATE: begin

				if( current_burst_count_plus1 >= reg_this_write_request_burst_count ) begin		// this is the last burst for this TLP
				
					state <= INITIAL_HEADER_1_0_STATE;											// *** waitrequest deasserts here ***
					current_burst_count <= 0;
					
				end else begin																	// this is NOT the last burst for this TLP
				
					current_burst_count <= current_burst_count + 1;								// *** waitrequest deasserts here ***
					
				end
			end
			default: begin
			
				state <= INITIAL_HEADER_1_0_STATE;
			end
			endcase
		end
	end
end

//
// TAG management
//
assign avm_tagreq_address = { current_tag_value, 2'b00 };
	
assign avm_tagreq_byteenable = 4'hF;

assign avm_tagreq_write = ( current_tag_value == next_tag_value ) ? ( 1'b0 ) : ( 1'b1 );

assign avm_tagreq_writedata = 32'h0;

assign next_tag_value_plus1 = next_tag_value + 1;

assign next_tag_available = ( next_tag_value_plus1 == current_tag_value ) ? ( 1'b0 ) : ( 1'b1 );

assign advance_next_tag =	( ( state == INITIAL_HEADER_1_0_STATE ) || ( state == CONTINUE_HEADER_1_0_STATE ) ) &&
							( request_outpipe_ready & request_outpipe_valid & p1_avs_s0_read );

assign advance_current_tag = asi_rddatasnk_ready & asi_rddatasnk_valid & asi_rddatasnk_endofpacket;

always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin

		current_tag_value <= 0;
		next_tag_value <= 0;
		tag_allocate_error <= 0;
		tag_overrun_error <= 0;

	end else begin
	
		if( advance_next_tag  ) begin
		
			if( next_tag_available ) begin
			
				next_tag_value <= next_tag_value + 1;
				
			end else begin
				
				tag_allocate_error <= 1;
			end
		end

		if( advance_current_tag  ) begin
		
			current_tag_value <= current_tag_value + 1;

			if( current_tag_value == next_tag_value ) begin
			
				tag_overrun_error <= 1;
			end
		end
	end
end

//
// Input Pipeline
//
slave2tlp_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 1 + 1 + 10 + 10 + SLAVE_ADDRESS_WIDTH + 8 + 10 + 1 + 1 + 64 ) ) in_pipe_1 (
    .clk            ( clk ),
    .reset_n        ( ~reset ),
    .in_ready       ( in_pipe_1_in_ready ),
    .in_valid       ( avs_s0_read | avs_s0_write ), 
    .in_payload     ( { more_than_one_write_request, more_than_one_read_request, allowed_write_burstcount_in_first_request, allowed_read_burstcount_in_first_request, avs_s0_address, avs_s0_byteenable, avs_s0_burstcount, avs_s0_read, avs_s0_write, avs_s0_writedata } ),
    .out_ready      ( waitrequest_inv ), 
    .out_valid      ( in_pipe_1_out_valid ), 
    .out_payload    ( { p1_more_than_one_write_request, p1_more_than_one_read_request, p1_allowed_write_burstcount_in_first_request, p1_allowed_read_burstcount_in_first_request, p1_avs_s0_address, p1_avs_s0_byteenable, p1_avs_s0_burstcount, p1_avs_s0_read, p1_avs_s0_write, p1_avs_s0_writedata } )
);

//
// Read Data output Pipeline
//
assign avs_s0_readdatavalid = readdata_outpipe_valid;

slave2tlp_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 64 ) ) readdata_outpipe (
    .clk            ( clk ),
    .reset_n        ( ~reset ),
    .in_ready       ( asi_rddatasnk_ready ),
    .in_valid       ( asi_rddatasnk_valid ), 
    .in_payload     ( asi_rddatasnk_data ),
    .out_ready      ( 1'b1 ), 
    .out_valid      ( readdata_outpipe_valid ), 
    .out_payload    ( avs_s0_readdata )
);

//
// Request output Pipeline
//
slave2tlp_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 5 ) ) eoplbe_outpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (eoplbe_outpipe_ready),
    .in_valid       (eoplbe_outpipe_valid), 
    .in_payload     ( { write_req_sub_1dw, write_req_lbe_bits } ),
    .out_ready      ( aso_eoplbesrc_ready ), 
    .out_valid      ( aso_eoplbesrc_valid ), 
    .out_payload    ( aso_eoplbesrc_data )
);

//
// Request output Pipeline
//
slave2tlp_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 66 ) ) request_outpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (request_outpipe_ready),
    .in_valid       (request_outpipe_valid), 
    .in_payload     ( { request_data, request_sop, request_eop } ),
    .out_ready      (aso_tlpsrc_ready), 
    .out_valid      (aso_tlpsrc_valid), 
    .out_payload    ( { aso_tlpsrc_data, aso_tlpsrc_startofpacket, aso_tlpsrc_endofpacket } )
);

// --------------------------------------------------
// Calculates the log2ceil of the input value
// --------------------------------------------------
function integer log2ceil;
	input integer val;
	integer i;

	begin
		i = 1;
		log2ceil = 0;

		while (i < val) begin
			log2ceil = log2ceil + 1;
			i = i << 1; 
		end
	end
endfunction

endmodule

//  --------------------------------------------------------------------------------
// | single buffered pipeline stage
//  --------------------------------------------------------------------------------
module slave2tlp_64b_1stage_pipeline  
#( parameter PAYLOAD_WIDTH = 8 )
 ( input                          clk,
   input                          reset_n, 
   output reg                     in_ready,
   input                          in_valid,   
   input      [PAYLOAD_WIDTH-1:0] in_payload,
   input                          out_ready,   
   output reg                     out_valid,
   output reg [PAYLOAD_WIDTH-1:0] out_payload      
 );
      
   always @* begin
     in_ready <= out_ready || ~out_valid;
   end
   
   always @ (negedge reset_n, posedge clk) begin
      if (!reset_n) begin
         out_valid <= 0;
         out_payload <= 0;
      end else begin
         if (in_valid) begin
           out_valid <= 1;
         end else if (out_ready) begin
           out_valid <= 0;
         end
         
         if(in_valid && in_ready) begin
            out_payload <= in_payload;
         end
      end
   end

endmodule
