//--------------------------------------------------------------------------
// master_bfm_tb.v 
// This test bench instantiates a master BFM and an onchip RAM
// as the slave DUT. The test starts with master BFM writing an
// amount of data to the slave according to the number of
// transactions which is defined by user.
//
// Next, master BFM starts reading back the data from slave.
// The read back data is compared with the desired data .
//
// The test case is defined in the test_input.v file while the
// result checking rountine is defined in test_result.v file
//--------------------------------------------------------------------------

// console messaging level
`define VERBOSITY VERBOSITY_INFO

//BFM hierachy
`define MSTR_BFM my_pcie_gen2x4_chaining_testbench.ep.test_sys_sopc_inst.the_mm_master_bfm_0.mm_master_bfm_0
`define PCIE_RP_BFM_DRVR my_pcie_gen2x4_chaining_testbench.drvr
`define SOPC_SYS my_pcie_gen2x4_chaining_testbench.ep.test_sys_sopc_inst

//test bench parameters
`define NUM_TRANS 5  //user-defined number of transactions to be performed
`define INDEX_ZERO 0  //burst cycle with index zero for non-bursting transactions

// BFM related parameters
`define AV_ADDRESS_W 32
`define AV_SYMBOL_W 8
`define AV_NUMSYMBOLS 8
`define AV_BURSTCOUNT_W 10

// derived parameters
`define AV_DATA_W (`AV_SYMBOL_W * `AV_NUMSYMBOLS)

// offset into PCIe memory space
`define PCIE_32BIT_TEST_ADDR_OFFSET (    0 * 1024 )
`define PCIE_64BIT_TEST_ADDR_OFFSET ( 1024 * 1024 )

// offset into SOPC system space
`define SLAVE_0_CSR_BASE ( 'h8000_0000 )

//-----------------------------------------------------------------------------
// Test Top Level begins here
//-----------------------------------------------------------------------------
module rsf_master_bfm_tb();

	//importing verbosity and avalon_mm packages
	import verbosity_pkg::*;
	import avalon_mm_pkg::*;

	// instantiate SOPC system module
	//  test_bench tb();  

	//local variables   
	Request_t command_request, response_request;
	reg [`AV_ADDRESS_W-1:0] command_addr, command_addr_base, response_addr;
	reg [`AV_DATA_W-1:0] command_data, command_data_mask, response_data;
	reg [`AV_NUMSYMBOLS-1:0] byte_enable, last_byte_enable, response_byte_enable;
	reg [`AV_BURSTCOUNT_W-1:0]burst_count;
	reg [`AV_BURSTCOUNT_W-1:0]burst_size, response_burst_size;
	reg [`AV_DATA_W-1:0] idle;  
	integer init_latency;
	reg [`AV_DATA_W-1:0] master_data_scoreboard [$];
	reg [`AV_DATA_W-1:0] master_mask_scoreboard [$];
	reg [`AV_DATA_W-1:0] expected_data;    
	reg [`AV_DATA_W-1:0] expected_mask;    
	integer failure = 0;
	integer checked = 0;
	integer response_count = 0;
	integer request_count = 0;
	integer response_index = 0;
	integer input_complete = 0;
	integer av_bfm_complete = 0;
	integer index_4k = 0;
	event start_test;
	event go_check;
	reg [`AV_DATA_W-1:0] burst_data;
	integer bursts_written = 0;
  
	//initialize the master BFM
	initial begin
		set_verbosity(`VERBOSITY);
		`MSTR_BFM.init();	
		`MSTR_BFM.set_response_timeout(5000);
		`MSTR_BFM.set_command_timeout(5000);
		
		$sformat(message, "");
		print(VERBOSITY_INFO, message);        								 
		$sformat(message, "Waiting for reset_n... ");
		print(VERBOSITY_INFO, message);        								 
		$sformat(message, "");
		print(VERBOSITY_INFO, message);        								 

		wait( `SOPC_SYS.reset_n == 1 );

		$sformat(message, "");
		print(VERBOSITY_INFO, message);        								 
		$sformat(message, "Waiting for run_avalon_bfm_now... ");
		print(VERBOSITY_INFO, message);        								 
		$sformat(message, "");
		print(VERBOSITY_INFO, message);        								 

		wait( `PCIE_RP_BFM_DRVR.run_avalon_bfm_now == 1 );

		$sformat(message, "");
		print(VERBOSITY_INFO, message);        								 
		$sformat(message, "Saw run_avalon_bfm_now... ");
		print(VERBOSITY_INFO, message);        								 
		$sformat(message, "");
		print(VERBOSITY_INFO, message);        								 

		-> start_test;
	end
    
	//check responses received by master BFM
	always @( posedge `MSTR_BFM.clk ) begin

		while (`MSTR_BFM.get_response_queue_size() > 0) begin	     

			`MSTR_BFM.pop_response();
			response_count++;

			//trigger event to check the response with expected data
			-> go_check;

			wait( checked == response_count );
		end
	end
  
	//simulation ends here
	//both write and read transactions do have response descriptors, so the number 
	//of responses received by master BFM is double of NUM_TRANS
	initial begin

		$sformat(message, "Waiting for command issue queue to start filling.");
		print(VERBOSITY_INFO, message);
		while ( `MSTR_BFM.get_command_issued_queue_size() == 0 )
		@( `MSTR_BFM.clk );

		$sformat(message, "Waiting for input_complete to be set.");
		print(VERBOSITY_INFO, message);
		wait( input_complete == 1);

		$sformat(message, "Waiting for all transactions to complete.");
		print(VERBOSITY_INFO, message);
		while ( `MSTR_BFM.all_transactions_complete() != 1 )
		@( `MSTR_BFM.clk );

		$sformat(message, "Waiting for all responses to be fetched.");
		print(VERBOSITY_INFO, message);
		while ( `MSTR_BFM.get_response_queue_size() != 0 )
		@( `MSTR_BFM.clk );

		wait( checked == request_count );

		$sformat(message, "Test has completed.");
		print(VERBOSITY_INFO, message);
		$sformat(message, "TEST_SUCCESS");
		print(VERBOSITY_INFO, message);

		av_bfm_complete = 1;

		wait( `SOPC_SYS.reset_n == 0 );
		$stop;
	end
    
	`include "rsf_test_input.v"
	`include "rsf_test_result.v"
  
	//----------------------------------------------------------------------------------
	// tasks
	//----------------------------------------------------------------------------------

	//this task sets the command descriptor for master BFM and push it to the queue
	task master_set_and_push_command;
		input Request_t request;
		input [`AV_ADDRESS_W-1:0] addr;
		input [`AV_DATA_W-1:0] data;
		input [`AV_NUMSYMBOLS-1:0] byte_enable;
		input [`AV_BURSTCOUNT_W-1:0]burst_count;
		input [`AV_BURSTCOUNT_W-1:0]burst_size;
		input [`AV_DATA_W-1:0] idle;
		input [31:0] init_latency;

		begin
		
			wait( `MSTR_BFM.get_command_issued_queue_size() < 32 )

			`MSTR_BFM.set_command_request(request);
			`MSTR_BFM.set_command_address(addr);    
			`MSTR_BFM.set_command_byte_enable(byte_enable,`INDEX_ZERO);
			`MSTR_BFM.set_command_burst_count(burst_count);    
			`MSTR_BFM.set_command_burst_size(burst_size);    
			`MSTR_BFM.set_command_idle(idle, `INDEX_ZERO);
			`MSTR_BFM.set_command_init_latency(init_latency);

			if (request == REQ_WRITE) begin
				`MSTR_BFM.set_command_data(data, `INDEX_ZERO);      
			end

			`MSTR_BFM.push_command();
			request_count++;
		end
	endtask
  
	//this task sets the command descriptor for master BFM and push it to the queue
	task master_set_and_push_command_write_burst;
		input [`AV_ADDRESS_W-1:0] addr;
		input [`AV_NUMSYMBOLS-1:0] byte_enable;
		input [`AV_NUMSYMBOLS-1:0] last_byte_enable;
		input [`AV_BURSTCOUNT_W-1:0]burst_count;
		input [`AV_BURSTCOUNT_W-1:0]burst_size;
		input [`AV_DATA_W-1:0] idle;
		input [31:0] init_latency;

		begin

			wait( `MSTR_BFM.get_command_issued_queue_size() < 32 )

			bursts_written = 0;
			`MSTR_BFM.set_command_request(REQ_WRITE);
			`MSTR_BFM.set_command_address(addr);    
			`MSTR_BFM.set_command_burst_count(burst_count);    
			`MSTR_BFM.set_command_burst_size(burst_size);    
			`MSTR_BFM.set_command_init_latency(init_latency);

			burst_data = 'h0706050403020100;
			`MSTR_BFM.set_command_byte_enable( byte_enable, bursts_written );
			`MSTR_BFM.set_command_data( burst_data, bursts_written );
			`MSTR_BFM.set_command_idle(idle, bursts_written);
			command_data_mask[  7 :  0 ] =	( byte_enable[0] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 15 :  8 ] =	( byte_enable[1] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 23 : 16 ] =	( byte_enable[2] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 31 : 24 ] =	( byte_enable[3] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 39 : 32 ] =	( byte_enable[4] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 47 : 40 ] =	( byte_enable[5] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 55 : 48 ] =	( byte_enable[6] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 63 : 56 ] =	( byte_enable[7] ) ? ( 8'hFF ) : ( 8'h00 );
			master_data_scoreboard.push_back( burst_data | ~command_data_mask );
			master_mask_scoreboard.push_back( 64'hFFFF_FFFF_FFFF_FFFF );
			bursts_written++;
			burst_data = burst_data + 'h0808080808080808;

			while( ( bursts_written + 1 ) < burst_size ) begin
				`MSTR_BFM.set_command_byte_enable( 'hFF, bursts_written );
				`MSTR_BFM.set_command_data( burst_data, bursts_written );
				`MSTR_BFM.set_command_idle(idle, bursts_written);
				command_data_mask = 64'hFFFF_FFFF_FFFF_FFFF;
				master_data_scoreboard.push_back( burst_data );
				master_mask_scoreboard.push_back( command_data_mask );
				bursts_written++;
				if( burst_data == 'hFFFEFDFCFBFAF9F8 ) begin
					burst_data = 'h0706050403020100;
				end else begin
					burst_data = burst_data + 'h0808080808080808;
				end
			end

			`MSTR_BFM.set_command_byte_enable( last_byte_enable, bursts_written );
			`MSTR_BFM.set_command_data( burst_data, bursts_written );
			`MSTR_BFM.set_command_idle(idle, bursts_written);
			command_data_mask[  7 :  0 ] =	( last_byte_enable[0] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 15 :  8 ] =	( last_byte_enable[1] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 23 : 16 ] =	( last_byte_enable[2] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 31 : 24 ] =	( last_byte_enable[3] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 39 : 32 ] =	( last_byte_enable[4] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 47 : 40 ] =	( last_byte_enable[5] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 55 : 48 ] =	( last_byte_enable[6] ) ? ( 8'hFF ) : ( 8'h00 );
			command_data_mask[ 63 : 56 ] =	( last_byte_enable[7] ) ? ( 8'hFF ) : ( 8'h00 );
			master_data_scoreboard.push_back( burst_data | ~command_data_mask );
			master_mask_scoreboard.push_back( 64'hFFFF_FFFF_FFFF_FFFF );

			`MSTR_BFM.push_command();
			request_count++;
		end
	endtask

endmodule
