//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module passes the read descriptors down towards the read master after metering a read credit
//	algorithm to prevent oversubscribing the read data FIFO.  The descriptor word count is used to
//	increase the credit count.
//
//	  input descriptor 64-bit: { reserved[ 15 : 0 ], channel[ 7 : 0 ], word_count[ 7 : 0 ], address[ 31 : 0 ] }
//
//	This module manages a read data credit scheme such that it does not oversubscribe the external read
//	data FIFO which is described by the READ_CREDIT_WIDTH parameter.  You should set READ_CREDIT_WIDTH
//	such that the READ_CREDIT_LIMIT macro is calculated properly by this formula: 
//		 READ_CREDIT_LIMIT = 2 ** ( READ_CREDIT_WIDTH - 1 )
//	READ_CREDIT_LIMIT should not exceed the total number of Avalon ST data words that the external read
//	data FIFO can store.
//
//	 256 word deep FIFO : 'h100 words :  9-bit credit width : 2 ^  8 = 256
//	 512 word deep FIFO : 'h200 words : 10-bit credit width : 2 ^  9 = 512
//	1024 word deep FIFO : 'h400 words : 11-bit credit width : 2 ^ 10 = 1024
//	2048 word deep FIFO : 'h800 words : 12-bit credit width : 2 ^ 11 = 2048
//	

module read_credit_gate_64b
#(
	parameter READ_CREDIT_WIDTH = 10
)
(

	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// Descriptor Input Interface - Avalon ST
	input  wire	[ 63 : 0 ]	asi_descsnk_data,
	output wire				asi_descsnk_ready,
	input  wire				asi_descsnk_valid,
	
	// Descriptor Output Interface - Avalon ST
	output wire	[ 63 : 0 ]	aso_descsrc_data,
	input  wire				aso_descsrc_ready,
	output wire				aso_descsrc_valid,

	// Read Credit Decrement Input Interface - Avalon ST
	input  wire 			asi_readcreditdec_data
	
);

localparam READ_CREDIT_LIMIT = 2 ** ( READ_CREDIT_WIDTH - 1 );

//
// Things we need to compute
//
reg  [ ( READ_CREDIT_WIDTH - 1 ) : 0] read_credits;
wire [  7 : 0 ]	credits_to_add;
wire [  7 : 0 ]	credits_to_sub;
wire			credit_available;
wire			in_ready;
wire			in_valid;

//
// Descriptor Elements
//
wire [ 7 : 0 ] word_count;
wire [ 7 : 0 ] channel;

assign word_count	= asi_descsnk_data[ 39 : 32 ];
assign channel		= asi_descsnk_data[ 47 : 40 ];

assign credit_available = ( ( word_count + read_credits ) > READ_CREDIT_LIMIT ) ? ( 1'b0 ) : ( 1'b1 );

assign asi_descsnk_ready	=	( channel == 0 ) ?
									( in_ready & credit_available ) : 
									( in_ready );

assign in_valid = asi_descsnk_valid & asi_descsnk_ready;

//
// Read Credit Managment
//
assign credits_to_add = ( in_valid & asi_descsnk_ready & ( channel == 0 ) ) ? ( word_count ) : ( 8'd0 );

assign credits_to_sub = ( asi_readcreditdec_data ) ? ( 8'd1 ) : ( 8'd0 );

always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin

		read_credits <= 0;
		
	end else begin

		read_credits <= read_credits + credits_to_add - credits_to_sub;
		
	end
end

//
// output Pipeline
//
read_credit_gate_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 64 ) ) outpipe (
    .clk            (clk),
    .reset_n        (~reset),
    .in_ready       (in_ready),
    .in_valid       (in_valid), 
    .in_payload     (asi_descsnk_data),
    .out_ready      (aso_descsrc_ready), 
    .out_valid      (aso_descsrc_valid), 
    .out_payload    (aso_descsrc_data)
);

endmodule

//  --------------------------------------------------------------------------------
// | single buffered pipeline stage
//  --------------------------------------------------------------------------------
module read_credit_gate_64b_1stage_pipeline  
#( parameter PAYLOAD_WIDTH = 8 )
 ( input                          clk,
   input                          reset_n, 
   output reg                     in_ready,
   input                          in_valid,   
   input      [PAYLOAD_WIDTH-1:0] in_payload,
   input                          out_ready,   
   output reg                     out_valid,
   output reg [PAYLOAD_WIDTH-1:0] out_payload      
 );
      
   always @* begin
     in_ready <= out_ready || ~out_valid;
   end
   
   always @ (negedge reset_n, posedge clk) begin
      if (!reset_n) begin
         out_valid <= 0;
         out_payload <= 0;
      end else begin
         if (in_valid) begin
           out_valid <= 1;
         end else if (out_ready) begin
           out_valid <= 0;
         end
         
         if(in_valid && in_ready) begin
            out_payload <= in_payload;
         end
      end
   end

endmodule
