//----------------------------------------------------------------------------------------------------
//-- (c)2010 Altera Corporation. All rights reserved.
//--
//-- Altera products are protected under numerous U.S. and foreign patents,
//-- maskwork rights, copyrights and other intellectual property laws.
//--
//-- This reference design file, and your use thereof, is subject to and governed
//-- by the terms and conditions of the applicable Altera Reference Design License
//-- Agreement (either as signed by you or found at www.altera.com).  By using
//-- this reference design file, you indicate your acceptance of such terms and
//-- conditions between you and Altera Corporation.  In the event that you do not
//-- agree with such terms and conditions, you may not use the reference design
//-- file and please promptly destroy any copies you have made.
//--
//-- This reference design file is being provided on an as-is basis and as an
//-- accommodation and therefore all warranties, representations or guarantees of
//-- any kind (whether express, implied or statutory) including, without
//-- limitation, warranties of merchantability, non-infringement, or fitness for
//-- a particular purpose, are specifically disclaimed.  By making this reference
//-- design file available, Altera expressly does not recommend, suggest or
//-- require that this reference design file be used in combination with any
//-- other product not provided by Altera.
//----------------------------------------------------------------------------------------------------

//	
//	This module receives Avalon master commands in two descriptors with the following format:
//
//	address descriptor 40-bit: { word_count[ 7 : 0 ], address[ 31 : 0 ] }
//	   data descriptor 16-bit: { channel[ 7 : 0 ], word_count[ 7 : 0 ] }
//
//	It uses the address descriptor to generate read requests thru it's Avalon Master interface, and it uses
//	the data descriptor to demark the return data with SOP and EOP and channelize the data.
//	

module read_master_64b
(
	// Clock Interface
	input  wire				clk,
	input  wire				reset,
	
	// Avalon Address Descriptor Input Interface - Avalon ST
	input  wire	[ 39 : 0 ]	asi_addrdesc_data,
	output wire				asi_addrdesc_ready,
	input  wire				asi_addrdesc_valid,

	// Avalon Data Descriptor Input Interface - Avalon ST
	input  wire	[ 15 : 0 ]	asi_datadesc_data,
	output wire				asi_datadesc_ready,
	input  wire				asi_datadesc_valid,

	// Avalon Master Read Data Output Interface - Avalon ST
	output wire	[ 63 : 0 ]	aso_rddatasrc_data,
	input  wire				aso_rddatasrc_ready,
	output wire				aso_rddatasrc_valid,
	output wire	[  7 : 0 ]	aso_rddatasrc_channel,
	output wire				aso_rddatasrc_startofpacket,
	output wire				aso_rddatasrc_endofpacket,

	// Master Interface - Avalon MM
	output wire	[ 31 : 0 ]	avm_m0_address,
	output wire	[  7 : 0 ]	avm_m0_byteenable,
	output wire	[  9 : 0 ]	avm_m0_burstcount,
	output wire				avm_m0_read,
	input  wire				avm_m0_waitrequest,
	input  wire	[ 63 : 0 ]	avm_m0_readdata,
	input  wire				avm_m0_readdatavalid
);

//
// Misc Variables
//
reg  [ 7 : 0 ] word_received_count;
wire [ 7 : 0 ] next_word_received_count;
wire           sop;
wire           eop;

//
// Master Command Input Elements
//
wire [ 31 : 0 ] avm_address;
wire [  7 : 0 ]	avm_burstcount;
wire [  7 : 0 ]	data_word_count;
wire [  7 : 0 ] data_channel;

assign { avm_burstcount, avm_address } = asi_addrdesc_data;
assign { data_channel, data_word_count } = asi_datadesc_data;

assign avm_m0_address = avm_address;
assign avm_m0_byteenable = 8'hFF;
assign avm_m0_burstcount = avm_burstcount;

assign avm_m0_read = asi_addrdesc_valid;

assign asi_addrdesc_ready = avm_m0_read & ~avm_m0_waitrequest;

assign asi_datadesc_ready = ( eop ) ? ( avm_m0_readdatavalid ) : ( 1'b0 );

assign next_word_received_count = word_received_count + 1;

assign sop = ( word_received_count == 0 ) ? ( 1'b1 ) : ( 1'b0 );

assign eop = ( next_word_received_count == data_word_count ) ? ( 1'b1 ) : ( 1'b0 );

always @ ( posedge clk or posedge reset ) begin
	if( reset ) begin
	
		word_received_count <= 0;
		
	end else begin
		if( avm_m0_readdatavalid ) begin
			if( eop ) begin
			
				word_received_count <= 0;
				
			end else begin

				word_received_count <= next_word_received_count;
			end
		end
	end
end

//
// readdata output Pipeline
//
read_master_64b_1stage_pipeline #( .PAYLOAD_WIDTH( 1 + 1 + 8 + 64 ) ) readdata_outpipe (
    .clk            ( clk ),
    .reset_n        ( ~reset ),
    .in_ready       ( ),
    .in_valid       ( avm_m0_readdatavalid ), 
    .in_payload     ( { sop, eop, data_channel, avm_m0_readdata } ),
    .out_ready      ( aso_rddatasrc_ready ), 
    .out_valid      ( aso_rddatasrc_valid ), 
    .out_payload    ( { aso_rddatasrc_startofpacket, aso_rddatasrc_endofpacket, aso_rddatasrc_channel, aso_rddatasrc_data } )
);

endmodule

//  --------------------------------------------------------------------------------
// | single buffered pipeline stage
//  --------------------------------------------------------------------------------
module read_master_64b_1stage_pipeline  
#( parameter PAYLOAD_WIDTH = 8 )
 ( input                          clk,
   input                          reset_n, 
   output reg                     in_ready,
   input                          in_valid,   
   input      [PAYLOAD_WIDTH-1:0] in_payload,
   input                          out_ready,   
   output reg                     out_valid,
   output reg [PAYLOAD_WIDTH-1:0] out_payload      
 );
      
   always @* begin
     in_ready <= out_ready || ~out_valid;
   end
   
   always @ (negedge reset_n, posedge clk) begin
      if (!reset_n) begin
         out_valid <= 0;
         out_payload <= 0;
      end else begin
         if (in_valid) begin
           out_valid <= 1;
         end else if (out_ready) begin
           out_valid <= 0;
         end
         
         if(in_valid && in_ready) begin
            out_payload <= in_payload;
         end
      end
   end

endmodule
